# -*- coding: utf-8 -*-
"""

    Agatta: Three-item analysis Python package
    Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

    Agatta is a set of tools in the cladistic framework to perform
    three-item analysis and associated operations in cladistics in python.

    https://github.com/vrineau/agatta

    This code is under license GNU GPLv3

"""

from fractions import Fraction
from itertools import combinations
from .ini import taxa_extraction
from .ini import character_extraction
from .ini import wrapper_character_extraction
from .analysis import standard_tripdec
from .analysis import rep_detector
from .analysis import del_replications_forest
from .analysis import main_tripdec
from .analysis import triplet_extraction
import csv
from os import path, devnull, remove
import sys
from pypdf import PdfMerger
from fractions import Fraction
from ete3 import Tree

def constrict(treelist, prefix=False, silent=False):
    """
    Compute the strict consensus from a list of rooted trees.

    Parameters
    ----------
    treelist : list
        List of ete3.Tree objects.
    prefix : str or bool, optional
        Prefix of the file prefix.constrict to be saved with the strict
        consensus in newick format. The default is False (no file saved).

    Returns
    -------
    constrict : ete3.Tree
        Strict consensus.

    """

    if not silent:
        print("Strict consensus computation")

    if rep_detector(dict.fromkeys(treelist, 1)):
        print("ERROR: Repeated leaves have been detected.\n" +
                         "Operation aborted.")
        sys.exit(1)

    for t in treelist:
        for leaf in t.get_leaf_names():
            if not leaf in treelist[0].get_leaf_names():
                print("ERROR: All trees do not have the same " +
                         "length.\nOperation aborted.")
                sys.exit(1)


    constrict = treelist[0]  # base tree for the strict consensus computation
    del treelist[0]
    kill_node_list = []  # list of nodes to delete at the end

    if treelist:
        for node in constrict.traverse(strategy="preorder"):  # for each node
            if not node.is_leaf() and not node.is_root():  # check if exists in all
                for tree_compare in treelist:  # for each tree
                    node_find = False
                    for node_tree_compare in tree_compare.traverse(
                            strategy="preorder"):
                        set1 = set([leaf.name for leaf in node.get_leaves()])
                        set2 = set([leaf.name for leaf in
                                    node_tree_compare.get_leaves()])
                        if set1 == set2:
                            node_find = True
                            break
                    if node_find == False:
                        kill_node_list.append(node)
                        break

        for kill_node in kill_node_list:  # del nodes not present in all trees
            kill_node.delete()

    if prefix:
        with open(prefix+".constrict", "w") as constrictfile:
            constrictfile.write(constrict.write(format=9)+"\n")

    if not silent:
        print("Strict consensus computed")

    return constrict


def rcc(treelist, prefix=False, verbose=False):
    """
    Compute the reduced cladistic consensus (Wilkinson, 1994, 1995) from a 
    list of rooted trees.

          Wilkinson, M. (1994). Common cladistic information and its consensus
          representation: reduced Adams and reduced cladistic consensus trees
          and profiles. Systematic Biology, 43(3), 343-368.
          
          Wilkinson, M. (1995). More on Reduced Consensus Methods. Systematic 
          Biology, 44(3). 


    Parameters
    ----------
    treelist : list
        List of ete3.Tree objects.
    prefix : str or bool, optional
        Prefix of the file prefix.constrict to be saved with the strict
        consensus in newick format. The default is False (no file saved).
    verbose : bool, optional.
        Verbose mode if True. Default is False.

    Returns
    -------
    profile : list
        List containing the strict consensus and all trees from the profile.

    """

    def nts_to_trees(cardinal, in_list):
        """
        Compute a ete3 tree from a set of n-taxon statements

        """

        ntstree = Tree()
        ntstree.add_features(taxa_content = cardinal)

        for nts in sorted(in_list, key=lambda x: len(x), reverse=True):
            for node_iter in ntstree.traverse(strategy="postorder"):

                # set of taxa from all the direct descendant nodes
                taxa_set = set()
                for node_child1 in node_iter.get_children():
                    taxa_set = taxa_set | node_child1.taxa_content

                # if place to add child node, check all nts by size and add
                if len(node_iter.taxa_content - taxa_set) > 1 and (
                        (node_iter.taxa_content - taxa_set) >= nts or (
                            node_iter.taxa_content - taxa_set) == nts):
                    node_iter.add_child().add_features(taxa_content = nts)
                    break

        #label nodes: paralog, ortholog, apical
        for node in ntstree.traverse(strategy="preorder"):
            if [node1 for node1 in node.get_children() if node1.taxa_content]:
                taxa_set = set()
                for node_child2 in node.get_children():
                    taxa_set = taxa_set | node_child2.taxa_content

                for taxa in node.taxa_content - taxa_set:
                    node.add_child(name=taxa).add_features(
                        taxa_content = set())  # add leaf for each of the set

            elif node.taxa_content:  # if no child node
                for taxa in node.taxa_content:
                    node.add_child(name=taxa).add_features(
                        taxa_content = set()) # add leaf for each of the set

        return ntstree


    print("Reduced cladistic consensus (RCC) computation")

    if rep_detector(dict.fromkeys(treelist, 1)):
        print("ERROR: Repeated leaves have been detected.\n" +
                         "Operation aborted.")
        sys.exit(1)
    else:
        for t in treelist:
            for leaf in t.get_leaf_names():
                if not leaf in treelist[0].get_leaf_names():
                    print("ERROR: All trees do not have the same " +
                             "length.\nOperation aborted.")
                    sys.exit(1)

    # check size of trees
    cardinal_list = []
    for treel in treelist:
        leaf_cptr = 0
        for node in treel.traverse(strategy="preorder"):
            if node.is_leaf() == True:
                leaf_cptr += 1
        cardinal_list.append(leaf_cptr)

    cardinal_set = set(cardinal_list)

    if len(cardinal_set) > 1:
        print("ERROR: Trees must all have the same leaf set." +
                       "\nOperation aborted.")
        sys.exit(1)

    # build list of components per tree
    component_trees = []

    for treel in treelist:
        component_trees.append([])
        for node in treel.traverse(strategy="preorder"):
            if not node.is_leaf() and not node.is_root():
                component_trees[-1].append(
                    [set([leaf.name for leaf in node.get_leaves()]), set(
                        [leaf.name for leaf in treel.get_leaves()]) - set(
                            [leaf.name for leaf in node.get_leaves()])])

    # build intersection list
    nts_intersect = []

    for comp1 in component_trees[0]:
        for comp2 in component_trees[1]:
            if len(comp1[0] & comp2[0]) > 1 and len(comp1[1] & comp2[1]) > 0:
                nts_intersect.append([comp1[0] & comp2[0],comp1[1] & comp2[1]])

    del component_trees[1]  # del trees from component_trees after transfert
    del component_trees[0]  #idem

    #redundancy deletion
    del_nts_intersect = []  # list where to add nts to delete
    for nts1, nts2 in combinations(nts_intersect, 2):
        if (nts1[0] <= nts2[0] and (
                nts1[1] <= nts2[1] or nts1[1] == nts2[1])) and (
                nts1 not in del_nts_intersect):
            del_nts_intersect.append(nts1)

        elif (nts1[0] >= nts2[0] and (
                nts1[1] >= nts2[1] or nts1[1] == nts2[1])) and (
                nts2 not in del_nts_intersect):
            del_nts_intersect.append(nts2)

    for del_nts in del_nts_intersect:  # del redundant nts
        while del_nts in nts_intersect:
            nts_intersect.remove(del_nts)

    #intersection of each set of components per tree with nts_intersect
    while len(component_trees) > 0:
        nts_intersect_new = []
        for comp1 in component_trees[-1]:  # pile
            for comp2 in nts_intersect:
                if len(comp1[0] & comp2[0]) > 1 and len(
                        comp1[1] & comp2[1]) > 0:
                    nts_intersect_new.append(
                        [comp1[0] & comp2[0],comp1[1] & comp2[1]])

        nts_intersect = nts_intersect_new

        del component_trees[-1]

        #redundancy deletion
        del_nts_intersect = []
        doubles_nts_intersect = []
        for nts1, nts2 in combinations(nts_intersect, 2):
            if nts1 == nts2:
                doubles_nts_intersect.append(nts1)

            elif (nts1[0] <= nts2[0] and (
                    nts1[1] <= nts2[1] or nts1[1] == nts2[1])) and (
                    nts1 not in del_nts_intersect):
                del_nts_intersect.append(nts1)

            elif (nts1[0] >= nts2[0] and (
                    nts1[1] >= nts2[1] or nts1[1] == nts2[1])) and (
                        nts2 not in del_nts_intersect):
                del_nts_intersect.append(nts2)

        for del_nts in doubles_nts_intersect:
            while nts_intersect.count(del_nts) != 1:
                nts_intersect.remove(del_nts)

        for del_nts in del_nts_intersect:
            while del_nts in nts_intersect:
                nts_intersect.remove(del_nts)

        if len(nts_intersect) == 0:  # if nts_intersect empyt, star tree
            break

    # gather nts in lists with common cardinal
    nts_groups = {}
    for nts in nts_intersect:
        nts.append(frozenset(nts[0]|nts[1]))
        nts_groups[frozenset(nts[0]|nts[1])] = []

    for nts in nts_intersect:
        nts_groups[nts[2]].append(nts[0])

    for nts in nts_intersect:
        for cardinal, in_list in nts_groups.items():
            if nts[2] != cardinal and nts[2] >= cardinal and len(
                    nts[0] & set(cardinal)) > 1:
                nts_groups[cardinal].append(nts[0] & set(cardinal))
                
    # for each pair of trees: second step from Wilkinson 1995     
    for card1, card2 in combinations(nts_groups, 2):
        
        # cardinal computation
        newcard = frozenset(card1 & card2)
        ntslist = []

        # new nts list computation
        if len(newcard) > 3 and newcard != card1 and newcard != card2:
            for nts in nts_groups[card1] + nts_groups[card2]:
                
                newnts = set(nts & newcard)
                
                if newnts not in ntslist and len(newnts) > 1:
                    ntslist.append(newnts)
                    
        # check redundancy and validate
        if (len(ntslist) > len(nts_groups[card1])) and (
                len(ntslist) > len(nts_groups[card2])):
            nts_groups[newcard] = ntslist

    #merging all nts into trees
    profile = []

    for cardinal, in_list in nts_groups.items():  #for each tree
        tree_profile = nts_to_trees(cardinal, in_list)
        if not len(tree_profile) == leaf_cptr:
            profile.append(tree_profile)

    profile.append(constrict(treelist, prefix=False, silent=True))

    #sort trees by size
    def get_len(in_set):
        return len(in_set)

    profile = sorted(profile, key=get_len, reverse=True)  # sort by size

    if verbose:
        if not cardinal_list[0] in [len(profile_tree)
                                    for profile_tree in profile]:
            print("The strict consensus is not informative")

        #affichage des résultats
        if len(nts_intersect) == 0:
            print("RCC profile empty: no common phylogenetic information")
        else:
            for tree_profile in profile:
                print(tree_profile)

    if prefix:
        with open(prefix+".rcc", "w") as rccfile:
            first_line = True
            i = 1
            for profiletree in profile:
                if first_line:
                    rccfile.write("Strict consensus:    "
                                  + profiletree.write(format=9)+"\n")
                    first_line = False
                else:
                    rccfile.write("Profile tree " + str(i) + ":    "
                                  + profiletree.write(format=9) + "\n")
                    i += 1

    print("RCC of {} trees computed".format(str(len(profile))))

    return profile


def RI_path(cladopath, charpaths, taxarep1=False, taxarep2=False, 
       method="TMS", weighting="FW", prefix=False, verbose=False):
    """
    Compute the retention index for hierarchical characters (Kitching et al., 
    1998) in the three-item analysis framework from two files.
    The function write an output file in which is a global retention index
    of the analysis and a retention index for each hierarchical character tree.

          Kitching, I. J., Forey, P., Humphries, C., & Williams, D. (1998).
          Cladistics: the theory and practice of parsimony analysis.
          Oxford University Press.

    Parameters
    ----------
    cladopath : str
        Path to a newick file containing the cladogram. The tree must be the
        optimal cladogram obtained from the cladistic analysis of character
        trees stored in charpath.
    charpaths : list of str
        List of paths to files containing character trees in newick or 
        hmatrix format.
    taxarep1 : str, optional
        DESCRIPTION. The default is False.
    taxarep2 : str, optional
        DESCRIPTION. The default is False.
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS" for removing polymorphism. 
        The default is "TMS".
    weighting : str, optional
        Weighting scheme to use between:
                                 * FW (Fractional weighting from
                                       Rineau et al. 2021)
                                 * FWNL (Fractional weighting from
                                       Nelson and Ladiges),
                                 * UW (Uniform weighting from
                                       Nelson and Ladiges 1992),
                                 * MW (Minimal Weighting),
                                 * AW (Additive Weighting),
                                 * NW (No Weighting).
        The default is "FW".
    prefix : str, optional
        Prefix of the saving file. The complete path can be
        used. The default is False (no file saved).

    Returns
    -------
    RI_char_dict : dict.
        Dictionary with character identifiers and global and per character RI.

    """

    print("Loading cladogram")

    cladopath = path.expanduser(cladopath)
    charpaths = [path.expanduser(charpath) for charpath in charpaths] 

    cladogram_dict = character_extraction(cladopath, taxarep1, verbose=False)

    print("Cladogram loaded")
    print("Loading character trees")

    character_dict = wrapper_character_extraction(charpaths, 
                                                  taxarep2,
                                                  prefix,
                                                  verbose=False)
        
    # Calls RI function
    RI_char_dict = RI(cladogram_dict, character_dict, taxarep1, taxarep2, 
           method, weighting, prefix, verbose)
    
    return RI_char_dict
        

def RI(cladogram_dict, character_dict, taxarep1=False, taxarep2=False, 
       method="TMS", weighting="FW", prefix=False, verbose=False):
    """
    Compute the retention index for hierarchical characters
    (Kitching et al., 1998) in the three-item analysis framework.
    The function write an output file in which is a global retention index
    of the analysis and a retention index for each hierarchical character tree.

          Kitching, I. J., Forey, P., Humphries, C., & Williams, D. (1998).
          Cladistics: the theory and practice of parsimony analysis.
          Oxford University Press.

    Parameters
    ----------
    cladogram_dict : dict
        Dictionary containing one newick tree (ete3 Tree objects) as key.
        The trees is generally the optimal tree or the strict consensus.
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
        The trees are generally the initial characters.
    taxarep1 : str, optional
        DESCRIPTION. The default is False.
    taxarep2 : str, optional
        DESCRIPTION. The default is False.
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS" for removing polymorphism. 
        The default is "TMS".
    weighting : str, optional
        Weighting scheme to use between:
                                 * FW (Fractional weighting from
                                       Rineau et al. 2021)
                                 * FWNL (Fractional weighting from
                                       Nelson and Ladiges),
                                 * UW (Uniform weighting from
                                       Nelson and Ladiges 1992),
                                 * MW (Minimal Weighting),
                                 * AW (Additive Weighting),
                                 * NW (No Weighting).
        The default is "FW". Note that retention index by character state 
        in FW mode is equivalent to FWNL.
    prefix : str, optional
        Prefix of the saving file. The complete path can be
        used. The default is False (no file saved).

    Returns
    -------
    RI_char_dict : dict.
        Dictionary with character identifiers and global and per character RI.

    """

    # remove automatically repetitions if detected (user message printed)
    if rep_detector(character_dict):
        character_dict = del_replications_forest(character_dict,
                                                 method=method,
                                                 prefix=prefix,
                                                 verbose=verbose)

    print("Computing retention index")

    str_character_dict = {}
    for t, i in character_dict.items():
        str_character_dict[t] = str(i)
        
        for leaf in t.get_leaf_names():
            if not leaf in list(cladogram_dict.keys())[0].get_leaf_names():
                print("ERROR: Character tree " + str(i)
                               + ": leaf set must be equal"
                               + " or a subset of the cladogram\n"
                               + "leaf set. Operation aborted.")
                sys.exit(1)

    c_triplet_dict = standard_tripdec(cladogram_dict,
                                      weighting,
                                      dec_detail=False,
                                      prefix=False,
                                      verbose=False)
    RI_tot = [0, 0]
    RI_char_dict = {}
    RI_char_dict_num = {}
    RI_char_dict_denom = {}

    # computation of the global RI in the same time as RI per character
    for chartree, keys in str_character_dict.items():
        RI_char_dict[keys] = 0
        RI_char_dict_num[keys] = 0
        RI_char_dict_denom[keys] = 0
        triplet_dict = standard_tripdec({chartree: keys},
                                        weighting,
                                        dec_detail=False,
                                        prefix=False,
                                        verbose=False)

        for triplet1, FW in triplet_dict.items():  # for each triplet in a char
            if triplet1 in c_triplet_dict:  # if triplet is in cladogram, add

                #RI numerator
                RI_char_dict_num[keys] += FW  # per character
                RI_tot[0] += FW  # total RI

            #RI denominator
            RI_char_dict_denom[keys] += FW
            RI_tot[1] += FW


    # computation of RI per character state
    # This functionality is only able when using hmatrix, which gives specific 
    # codes for each character state.
    
    for chartree, keys in str_character_dict.items():
        
        for node in chartree.traverse(strategy="preorder"):
            if node.is_leaf() == False and node.is_root() == False:
                try: 
                    charstate_count = node.charstate_name

                    # compute triplets for this single state
                    charstate = chartree.copy('cpickle')
                    for n in charstate.traverse(strategy="preorder"):
                        if n.is_leaf() == False and n.is_root() == False:
                            try:
                                if not n.charstate_name == charstate_count:
                                    n.delete()
                            except AttributeError:
                                n.delete()
                                
                    keystate = str(keys.split('.')[0]) + "_" + str(
                                                            charstate_count)
                    RI_char_dict[keystate] = 0
                    RI_char_dict_num[keystate] = 0
                    RI_char_dict_denom[keystate] = 0
                    
                    state_triplet_dict = standard_tripdec({charstate: keys},
                                                    weighting,
                                                    dec_detail=False,
                                                    prefix=False,
                                                    verbose=False)

                    for triplet1, FW in state_triplet_dict.items():  
                        if triplet1 in c_triplet_dict:
            
                            #RI numerator
                            RI_char_dict_num[keystate] += FW  # per state
            
                        #RI denominator
                        RI_char_dict_denom[keystate] += FW
            
                except AttributeError:
                    pass
    
    # merge all in RI_char_dict fractions dict and remove zero fractions
    for keys, values in RI_char_dict_num.items():
        if RI_char_dict_denom[keys] == 0:
            RI_char_dict[keys] = [0, 1]
        else:
            RI_char_dict[keys] = [RI_char_dict_num[keys],
                                    RI_char_dict_denom[keys]]

    # computation of ri when free-paralogy subtrees (poly)
    for keys, values in str_character_dict.items(): # build RI for entire chars
        if '.' in str(values):
            if not values.split('.')[0] in RI_char_dict:
                RI_char_dict[values.split('.')[0]] = [0, 0]

            RI_char_dict[values.split('.')[0]][0] += RI_char_dict[values][0]
            RI_char_dict[values.split('.')[0]][1] += RI_char_dict[values][1]

    if RI_tot[1] == 0:
        RI_char_dict["Total"] = [0, 1]  # add global RI
    else:
        RI_char_dict["Total"] = [RI_tot[0], RI_tot[1]]

    print("Retention index computed")

    # computation of ri per state (results in FW are not equivalent between 
    # states and characters because of the correction (Rineau et al. 2021).
    #
    # Rineau, V., Zaragüeta, R., & Bardin, J. (2021). Information content 
    # of trees: three-taxon statements, inference rules and dependency. 
    # Biological Journal of the Linnean Society, 133(4), 1152-1170.) 

    # Write file
    if prefix:
        with open(prefix+".ri", "w") as RI_file:
            
            def formatNumber(num): # remove zero after float
                if num % 1 == 0:
                    return int(num)
                else:
                    return num

            def writeri(keys, values):
                RI_string0 = "[" + str(keys) + "]"
                RI_string1 = str(formatNumber(round((float(Fraction(values[0],
                                                       values[1]))*100), 2)))
                RI_string2 = str(formatNumber(round((float(values[0])), 2)))
                RI_string3 = str(formatNumber(round((float(values[1])), 2)))
                
                RI_string = [RI_string0, RI_string1, RI_string2, RI_string3]
                
                print('\t'.join(RI_string).expandtabs(10))
                RI_file.write((' '.join(RI_string) + "\n").expandtabs(10))
            
            RI_string = ['Chars', 'RI', 'Retained', 'Total']
            
            print('\n' + '\t'.join(RI_string).expandtabs(10))
            RI_file.write((' '.join(RI_string) + "\n").expandtabs(10))
            
            for keys, values in RI_char_dict.items():  # IR characters
                if not '.' in keys and not '_' in keys:
                    writeri(keys, values)
                    
            RI_string = ['\nStates', 'RI', 'Retained', 'Total']
            
            print('\t'.join(RI_string).expandtabs(10))
            RI_file.write((' '.join(RI_string) + "\n").expandtabs(10))
                    
            statescodespresence = False
            for keys, values in RI_char_dict.items():  # IR states
                if not '.' in keys and '_' in keys:
                    writeri(keys, values)
                    statescodespresence = True
                    
            if not statescodespresence:
                info_string = ("NA (hmatrix mandatory to set states codes)")
                print(info_string)
                RI_file.write(info_string + "\n")

            RI_string = ['\nSubtrees', 'RI', 'Retained', 'Total']
            
            print('\t'.join(RI_string).expandtabs(10))
            RI_file.write((' '.join(RI_string) + "\n").expandtabs(10))

            polypresence = False
            for keys, values in RI_char_dict.items():  # IR subtrees
                if '.' in keys and not '_' in keys:
                    writeri(keys, values)
                    polypresence = True

            if not polypresence:
                info_string = "NA\n"
                print(info_string)
                RI_file.write(info_string)
                
    # Codes for RI_char_dict
    #
    # 1 -   character
    # 1.1 - character subtree coming from FPLSA
    # 1_1 - character 1, state 1
    #
    
    RI_char_dict_return = {}
    for keys, values in RI_char_dict.items():
        RI_char_dict_return[keys] = Fraction(values[0], values[1])
    
    return RI_char_dict_return


def triplet_distance(t1, t2, prefix=False, method="TMS", weighting="FW", 
                     silent=False, verbose=False):
    """
    Compute several metrics between two hierarchical trees including the 
    inter-tree retention index (ITRI; Grand et al. 2014, Rineau et al. 2015, 
    which is an asymmetric distance between trees) and the triplet distance.
    Can be used also in the specific case of comparing a reference tree to 
    another tree.

    Grand, A., Zaragüeta-Bagils, R., Velez, L. M., & Ung, V. (2014).
    A cladistic re-analysis of the Gadiformes (Teleostei, Paracanthopterygii)
    using three-item analysis. Zootaxa, 3889(4), 525-552.
    Rineau, V., Grand, A., Zaragüeta i Bagils, R., & Laurin, M. (2015). 
    Experimental systematics : Sensitivity of cladistic methods to 
    polarization and character ordering schemes. 
    Contributions to Zoology, 84(2), 129‑148.


    Parameters
    ----------
    t1 : dict
        Dictionary containing one newick tree (ete3 Tree object) as key.
        If one want comparing a tree to a reference tree, t1 is the 
        reference tree.
    t2 : dict
        Dictionary containing one newick tree (ete3 Tree object) as key.
        If one want comparing a tree to a reference tree, t2 is the 
        tree one wants to compare to the reference.
    prefix : str
        Prefix of the text file to be saved containing the resulting ITRI.
        The complete path can be used. The default is False (no file saved).
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS" for removing polymorphism. 
        The default is "TMS".
    weighting : str, optional
        Weighting scheme to use between:
                                 * FW (Fractional weighting from
                                       Rineau et al. 2021)
                                 * FWNL (Fractional weighting from
                                       Nelson and Ladiges),
                                 * UW (Uniform weighting from
                                       Nelson and Ladiges 1992),
                                 * MW (Minimal Weighting),
                                 * AW (Additive Weighting),
                                 * NW (No Weighting).
        The default is "FW".
    Returns
    -------
    float
        precision. ITRI(t1,t2) (Precision: (t1∩t2)/t2)) 
        Computing the ITRI{t1,t2}, i.e. proportion of the FW of the  
        3is of t1 that are also present in t2. If t1 is a reference tree, 
        ITRI12 is the Precision: proportion of triplets from the reconstructed 
        tree that are true (equivalent to power in Rineau et al. 2015).
    float
        recall. ITRI(t2,t1) (Recall: (t1∩t2)/t1))
        Computing the ITRI{t1,t2}, i.e. proportion of the FW of the 3is 
        of t2 that are also present in t1. If t1 is a reference tree, ITRI12 is
        the recall.: proportion of triplets from the true tree that have been
        retrieved (inverse of artefact in Rineau et al. 2015).
    float
        tripdistance. Harmonic mean of ITRI(t2,t1) and ITRI(t1,t2).
        Harmonic mean between the two asymmetric ITRI values
        computed. Gives a triplet distance between two trees.
        If t1 is a reference tree, it is the F-score (harmonic mean of 
        precision and recall).

    """
    
    def triplet_incongruence_detector(triplet, tripdic):
        "Test if a triplet is incompatible with the triplet set (True), or"
        "absent but compatible (False). The triplet must not be in the set"
        
        for t in tripdic:
            t1set = t.out_taxa.union(t.in_taxa)
            t2set = triplet.out_taxa.union(triplet.in_taxa)
            if t1set == t2set:
                return True
            
        return False

    # remove automatically repetitions if detected (user message printed)
    if rep_detector({t1: 0}):
        t1 = del_replications_forest({t1: 0},
                                                 method=method,
                                                 prefix=prefix,
                                                 verbose=verbose).items()[0]

    if rep_detector({t2: 0}):
        t2 = del_replications_forest({t2: 0},
                                                 method=method,
                                                 prefix=prefix,
                                                 verbose=verbose).items()[0]

    tt_tripdic = standard_tripdec({t1: 0},
                                      weighting,
                                      dec_detail=False,
                                      prefix=False,
                                      verbose=False)

    rt_tripdic = standard_tripdec({t2: 0},
                                      weighting,
                                      dec_detail=False,
                                      prefix=False,
                                      verbose=False)
    

            

    RI_reconstructed_tree = Fraction(0, 1)  # total reconstruted tree
    RI_true_tree = Fraction(0, 1)  # total true tree
    RI_intersect_reconstructed_tree = Fraction(0, 1)
    RI_intersect_true_tree = Fraction(0, 1)
    false_positives = Fraction(0, 1)
    false_negatives = Fraction(0, 1)
    incongruent_triplets_t1 = Fraction(0, 1)
    incongruent_triplets_t2 = Fraction(0, 1)
    neutral_triplets_t1 = Fraction(0, 1)
    neutral_triplets_t2 = Fraction(0, 1)
    
    for trip, FW in rt_tripdic.items():
        if trip in tt_tripdic:
            RI_intersect_reconstructed_tree += FW
        else:
            if triplet_incongruence_detector(trip, tt_tripdic):
                incongruent_triplets_t1 += FW
            else:
                neutral_triplets_t1 += FW
            false_positives += FW
        RI_reconstructed_tree += FW

    for trip, FW in tt_tripdic.items():
        if trip in rt_tripdic:
            RI_intersect_true_tree += FW
        else:
            if triplet_incongruence_detector(trip, rt_tripdic):
                incongruent_triplets_t2 += FW
            else:
                neutral_triplets_t2 += FW
            false_negatives += FW
        RI_true_tree += FW

    recall = float(Fraction(RI_intersect_true_tree, RI_true_tree))  # power

    try:
        precision = float(Fraction(RI_intersect_reconstructed_tree,
                               RI_reconstructed_tree))  # 1/artefact
    except ZeroDivisionError:
        precision  = 0

    # F1-score
    try:
        F1score = (2 * precision * recall) / (precision + recall)
    except ZeroDivisionError:
        F1score  = 0
        
    #Triplet distance sensu Dobson 1975 Dobson, A.J. (1975). Comparing the 
    # shapes of trees. In: Street, A.P., Wallis, W.D. (eds) Combinatorial 
    # Mathematics III. Lecture Notes in Mathematics, vol 452.
    tripdistance = abs(false_positives + false_negatives) / 2
    
    # Triplet congruence / similarity
    # Amount of common triplets between t1 and t2. It may be a 
    # different amount in t1 and t2 because of the fractionnal weighting
    tripcongruence = abs(RI_intersect_true_tree + 
                         RI_intersect_reconstructed_tree) / 2
    
    # Triplet congruence / similarity. Not the same as triplet distance, as
    # the triplet incongruence takes only into account the incompatible 
    # triplets, while triplet distance takes also into account triplet that are
    # different but still compatible (or more exactly, neutral)
    tripincongruence = incongruent_triplets_t1 + incongruent_triplets_t2

    tripneutral = neutral_triplets_t1 + neutral_triplets_t2
        
    # results list to display /save
    itristr = []

    itristr.append("Number of triplets in t1 (relevant elements) : "
                   + str(round(float(RI_true_tree), 3)))
    
    itristr.append("Number of triplets in t2 (retreived elements) : " 
                   + str(round(float(RI_reconstructed_tree), 3)))
    
    itristr.append("Number of triplets both in t1 and t2 " 
                   + "(true positives) : " 
                   + str(round(float(RI_intersect_true_tree), 3)) 
                   + " (t1 also in t2) "
                   + str(round(float(RI_intersect_reconstructed_tree), 3)) 
                   + " (t2 also in t1)")
    
    itristr.append("Number of triplets in t2 but not in t1 " 
                   + "(false positives) : " 
                   + str(round(float(false_positives), 3)))
    
    itristr.append("Number of triplets in t1 but not in t2 " 
                   + "(false negatives) : " 
                   + str(round(float(false_negatives), 3)))
    
    itristr.append("Triplet distance ((t2Δt1)/2)) : " 
                   + str(round(float(tripdistance), 3)))
    
    itristr.append("Triplet congruence (amount of triplets present in both" 
                   + " trees) (t2 in t1 + t1 in t2)/2)) : " 
                   + str(round(float(tripcongruence), 3)))

    itristr.append("Triplet incongruence (amount of triplets incompatibles " +
                   "between t1 and t2) : " 
                   + str(round(float(tripincongruence), 3)))

    itristr.append("Triplet neutral (amount of neutral triplets, i.e. " 
                   + "triplets compatibles which are not the same between " 
                   + "t1 and t2) : " 
                   + str(round(float(tripneutral), 3)))

    itristr.append("ITRI(t1,t2) (Precision: (t2 in t1)/t2)) : " 
                   + str(round(precision, 3)))
    
    itristr.append("ITRI(t2,t1) (Recall: (t1 in t2)/t1)) : " 
                   + str(round(recall, 3)))
    
    itristr.append("Triplet F1-score: " 
                   + "(2 * Precision * Recall) / " 
                   + "(Precision + Recall)) : " 
                   + str(round(F1score, 3))) 

    # save file
    if prefix:
        with open(prefix+".itri", "w") as itrifile:
            
            for line in itristr:
                itrifile.write(line + "\n")

    # display
    if not silent:
        for line in itristr:
            print(line)
            
    return (float(RI_true_tree), 
           float(RI_reconstructed_tree),
           float(RI_intersect_true_tree),
           float(RI_intersect_reconstructed_tree),
           float(false_positives), 
           float(false_negatives),
           float(tripdistance),  #Dobson 1975
           float(tripcongruence),
           float(tripincongruence),
           float(tripneutral),
           precision,  # ITRI(T1,T2)
           recall,     # ITRI(T2,T1)
           F1score)              
           

def cladogram_label(cladogram, clade_number_option, clade_type_option, 
                    pdf_files=False):
    """
    Label tree nodes for NRI and character state test.
    """
    
    #sometimes issues with pyqt installation
    if pdf_files:
        try:
            from ete3 import NodeStyle, TreeStyle, faces
            from ete3 import TextFace, COLOR_SCHEMES, CircleFace
            
        except ImportError:  # issue with ete3 imports
            print("A manual instal of PyQt5 is requested to use the --pdf "
                  + "functionality\nPlease install using 'pip install" +
                  " PyQt5' and rerun the command line")
            sys.exit(1)

    cladogram.ladderize()

    if clade_number_option == "yes":  # label cladogram nodes
        node_style_num_count = 1
        for node in cladogram.traverse(strategy="preorder"):
            if node.is_leaf() == False  and node.is_root() == False:
                if pdf_files:
                    style_num = TextFace(node_style_num_count, fsize=2)
                    node.add_face(style_num, column=1,
                                  position = "branch-bottom")
                node.add_feature("clade_label", node_style_num_count)
                node.name = node_style_num_count
                node_style_num_count += 1
            elif node.is_root() == True:
                node.add_feature("clade_label", "root")
            else:
                node.add_feature("clade_label", node.name)

    if clade_type_option == "yes":  # label cladogram node types
        for node in cladogram.traverse(strategy="postorder"):
            if node.is_leaf():
                node.add_feature("clade_type", "leaf")
            else:
                i = len([child_node for child_node in node.get_children()
                         if not child_node.is_leaf()])
                if i >= 2:
                    node.add_feature("clade_type", "paralog")
                elif i == 1:
                    node.add_feature("clade_type", "ortholog")
                elif i == 0:
                    node.add_feature("clade_type", "apical_node")

    syn_dict = {}
    for node_number in cladogram.traverse(strategy="postorder"):
        syn_dict[str(node_number.clade_label)] = {
            "accepted": [],"rejected": []}

    return cladogram, syn_dict


def NRI(cladopath, charpaths, taxarep1=False, taxarep2=False, prefix="rnri", 
        rnri_codes=False, weighting='FW', polymethod='TMS', totaltree=True,
        rescaled=True, pdf_files=False):
    """
    Compute the Nodal Retention Index and return a pdf with percentages of NRI
    for each character (or for each set of characters if rnri_codes=True) 
    plus a csv file and with the total amount of triplet weights by node.
    
    Nodal retention index: given a set of characters and a cladogram, gives the 
    amount of triplet from each character in agreement with each node. Allows 
    to see easily the relative support for each clade.

    Parameters
    ----------
    cladopath : str
        path of input file containing the cladogram (or consensus).
    charpaths : list of str
        list of path of input file containing characters in hmatrix or newick 
        format.
    taxarep1 : str, optional
        DESCRIPTION. The default is False.
    taxarep2 : str, optional
        DESCRIPTION. The default is False.
    prefix : str
        Prefix of the files prefix.nri (csv file + a newick tree with node 
        names), prefix.pdf (percentages) and prefix.total.pdf (total triplet
        weights) to be saved. 
        The default is 'rnri'.
    rnri_codes : bool
        By default, the prefix.pdf file will display a tree with a piechart for 
        each node representing the relative proportion of support from each 
        character. The user can 
    weighting : str, optional
        Weighting scheme to use between FW, FWNL, MW, AW, NW. 
        The default is 'FW'.
    polymethod : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS". The default is "TMS".
    totaltree : bool, optional
        If set to True, computes the prefix.total.pdf showing the support of 
        each node given the relative amounts of character triplets in 
        agreement. The default is True.
    rescaled : bool, optional
        If True, for prefix.total.pdf the absolute values will be be rescaled 
        to avoid the symetry effect inherent to triplet decomposition. 
        Because of this effect, a balanced character has a heavier weight
        than an inbalanced character.
        The default is True.

    Returns
    -------
    pie_percentages1 : dict
        dictionary with node names as keys and lists of percentages as values.
        The list correspond all percentages displayed in the piechart tree.

    """

    #sometimes issues with pyqt installation
    if pdf_files:
        try:
            from ete3 import NodeStyle, TreeStyle, faces
            from ete3 import TextFace, COLOR_SCHEMES, CircleFace
            
        except ImportError:  # issue with ete3 imports
            print("A manual instal of PyQt5 is requested to use the --pdf "
                  + "functionality\nPlease install using 'pip install" +
                  " PyQt5' and rerun the command line")
            sys.exit(1)
    
    # reading matrix cladogram csv
        
    character_dict = wrapper_character_extraction(charpaths, taxarep1, prefix,
                                                  verbose=False)
        
    cladogram_dict = character_extraction(cladopath, taxarep2, verbose=False)
    
    if rnri_codes:
        
        rnri_codes = path.expanduser(rnri_codes)
        
        with open(rnri_codes, "r") as rnri_file:
            try:
                dialect = csv.Sniffer().sniff(rnri_file.readline(),
                                              delimiters=[";","\t"," ","|"])
            except:
                print("ERROR: The table file '" + rnri_codes +
                                  "' is not adequately formated." +
                                  "\n Operation aborted.")
            else:
                no_exception = True
            if not 'no_exception' in locals():
                sys.exit(1)
        
        with open(rnri_codes, "r") as rnri_file:
        
            char_list_raw = csv.reader(rnri_file, delimiter=dialect.delimiter)
            rnri_codes = list(char_list_raw)
        
            if not len({len(l) for l in rnri_codes}) == 1:
                 print("ERROR: The table file '" + rnri_codes +
                                  "' is broken.\n Operation aborted.")
                 sys.exit(1)
    
    # default function use : each character is represented in the piechart
    else:
        rnri_codes = []
        for charname in character_dict.values():
            rnri_codes.append(['', 'character ' + str(charname)])
            
    # add names to internal nodes
    cladogram = list(cladogram_dict.keys())[0]
    cladogram.ladderize()
    pie_percentages = dict() # dict int node name / percentage list for piech.
    Node_weight = dict() # dict internal node name / percentage list for piech.
    Node_weight_rescaled = dict() # dict with rescaling denominator
    
    categories_rnri = []
    n = [categories_rnri.append(x) for x in [x[1] for x in rnri_codes] 
     if x not in categories_rnri]
    
    node_style_num_count = 1
    
    card = len(cladogram.get_leaf_names())
    rescaledtot = 0
    for node in cladogram.traverse(strategy="preorder"):
        if node.is_leaf() == False  and node.is_root() == False:
            cardn = len(node.get_leaf_names())
            rescaledtot += (cardn - 1)*(card - cardn)
        
    for node in cladogram.traverse(strategy="preorder"):
        if node.is_leaf() == False  and node.is_root() == False:
            node_style_num_countstr = str(node_style_num_count)
            
            if pdf_files:
                style_num = TextFace(node_style_num_countstr, fsize=2)            
                node.add_face(style_num, column=1,
                              position = "branch-bottom")
                
            node.add_feature("clade_label", node_style_num_countstr)
            node.name = node_style_num_countstr
            pie_percentages[node_style_num_countstr] = [
                Fraction(0,1)] * len(set([x[1] for x in rnri_codes]))
            Node_weight[node_style_num_countstr] = Fraction(0,1)
            node_style_num_count += 1
            
            # rescaling factor computation
            cardn = len(node.get_leaf_names())
            Node_weight_rescaled[
                node.name] = ((cardn - 1)*(card - cardn)) / rescaledtot
            
        else:
            node.add_feature("clade_label", False)
    
    # RNRI computation : decompose each character in triplets with weights
    # for each triplet : traverse the tree and check if agreement tree/triplet
    # if yes add value to pie_percentages
    
    print("Starting detailed triplet decomposition and FW computation for NRI")
    
    for character, carnb in character_dict.items():  # for each character
    
        print("Decomposition: character " + str(carnb))
        
        #blockPrint()
        triplist = main_tripdec({character : 1}, prefix=False, 
                                taxa_replacement=False, weighting=weighting, 
                                parallel='no', dec_detail=False, 
                                method=polymethod, verbose=False)
        #enablePrint()
            
        for triplet_rnri, FW in triplist.items():  # check if triplet is true
            
            in1name, in2name = list(triplet_rnri.in_taxa)
            
            in1 = cladogram.get_leaves_by_name(str(in1name))[0]
            in2 = cladogram.get_leaves_by_name(str(in2name))[0] 
            out = cladogram.get_leaves_by_name(list(
                triplet_rnri.out_taxa)[0])[0]
            
            in_node = set(Tree.get_leaf_names(in1.get_common_ancestor(in2)))
            nodeout1 = set(Tree.get_leaf_names(out.get_common_ancestor(in1)))
            nodeout2 = set(Tree.get_leaf_names(out.get_common_ancestor(in2)))
            
            # check if triplet is valid
            if (in_node.issubset(nodeout1) and in_node.issubset(nodeout2) and 
                in_node != nodeout1 and in_node != nodeout2):
                
                # add value in result list
                pie_percentages[in1.get_common_ancestor(in2).name][
                    categories_rnri.index(rnri_codes[int(carnb)-1][1])] += FW
                Node_weight[in1.get_common_ancestor(in2).name] += FW
    
    print("Triplet decomposition and FW computation completed")
    
    # round and compute percentages
    
    pie_percentages1 = dict()
    Node_weight2 = dict()
    
    for node_name, values_list in pie_percentages.items():
        
        total = sum(pie_percentages[node_name])
        pie_percentages1[node_name] = [
            round((x / total) * 100) for x in values_list]
        pie_percentages1[node_name].pop()
        totalsub = sum(pie_percentages1[node_name])
        pie_percentages1[node_name].append(100 - totalsub)
    
    for node_name, frac in Node_weight.items():
        Node_weight2[node_name] = round((float(frac)), 5)
    
    # display pie-percentages
    
    # Basic tree style
    if pdf_files:
    
        ts = TreeStyle()
        ts.show_leaf_name = True
        ts.margin_top = 10
        ts.margin_right = 10
        ts.margin_left = 10
        ts.margin_bottom = 10
        ts.show_scale = False
        colors=["#a6cee3",
                "#b2df8a",
                "#fb9a99",
                "#fdbf6f",
                "#cab2d6",
                "#ffff99",
                "#1f78b4",
                "#33a02c",
                "#e31a1c",
                "#ff7f00",
                "#6a3d9a",
                "#b15928",
                "#001b17",
                "#9dff47",
                "#794ffd",
                "#009243",
                "#01168f",
                "#a6ffb7",
                "#c2008e",
                "#044a00",
                "#af7eff",
                "#ce6500",
                "#0091b6",
                "#c71b00",
                "#000c2f",
                "#f7ffde",
                "#3e0012",
                "#ffcdc4",
                "#b40021",
                "#4b4800",
                "#a60052",
                "#986400"]
        
        #remove display nodes
        for n in cladogram.traverse():
           nstyle = NodeStyle()
           nstyle["fgcolor"] = "black"
           nstyle["size"] = 0
           n.set_style(nstyle)
           
        #Associate the PieChartFace only with internal nodes
        def pie_nodes(node):
            if not node.is_leaf() and not node.is_root():
                
                F= faces.PieChartFace(pie_percentages1[node.name],
                                      colors=colors, 
                                      width=50, height=50)
                F.border.width = None
                F.opacity = 0.6
                faces.add_face_to_node(F,node, 0, position="branch-right")
        
        ts.layout_fn = pie_nodes
        
        i = 0
        for cat in categories_rnri:
            if i > len(colors):
                print("WARNING: too many categories, "
                      +"piechart will be unreadable.")
            else:
                ts.legend.add_face(CircleFace(10, colors[i]), column=0)
                ts.legend.add_face(TextFace(cat, fsize=7), column=1)    
            
                i += 1 
            
        # if yes, save a pdf with node size reflecting overall triplet support
        if totaltree:
            
            ts2 = TreeStyle()
            ts2.show_leaf_name = True
            ts2.margin_top = 10
            ts2.margin_right = 10
            ts2.margin_left = 10
            ts2.margin_bottom = 10
            ts2.show_scale = False
            
            bubble_cladogram = cladogram.copy()
            
            def bubble_layout(node):
                if not node.is_leaf() and not node.is_root():
                    # Creates a sphere face whose size is proportion. to weight
                    C = CircleFace(radius=radius[node.name], 
                                   color="#6d87ed", style="sphere")
                    C.opacity = 0.6
                    faces.add_face_to_node(C, node, 0, position="float")
                    
            #réechelonner
            radius = dict()
            if rescaled:
                radiusr = dict()
                for nodename, val in Node_weight2.items():
                    radiusr[nodename] = (Node_weight2[nodename] / 
                                         Node_weight_rescaled[nodename])
                for nodename, val in radiusr.items():
                    radius[nodename] = (radiusr[nodename] / 
                                        max(radiusr.values()))*20
                    
            else:
                for nodename, val in Node_weight2.items():
                    radius[nodename] = (Node_weight2[nodename] / 
                                        max(Node_weight2.values()))*20
                    
            ts2.layout_fn = bubble_layout
            
            bubble_cladogram.render(prefix + ".total.pdf", tree_style=ts2)
        
        # save results
        cladogram.render(prefix + ".pdf", tree_style=ts)
    
    with open(prefix+".nri", "w") as results_file:
        
        results_file.write(' ')
        for e in categories_rnri:
            results_file.write(',' + e)
        results_file.write(",total,total_rescaled\n")  
        
        for carname, percentage_list in pie_percentages1.items():
            results_file.write('Node_' + carname)
            for p in percentage_list: 
                results_file.write(',' + str(p))
    
            radius = Node_weight2[carname]        
            radius_rescaled = Node_weight2[carname] / Node_weight_rescaled[
                carname]
                
            results_file.write("," + str(radius) + "," + 
                               str(radius_rescaled) + "\n")  
    
            
        results_file.write("\n" + cladogram.write(format=8) + "\n")
            
    print("Nodal Retention Index computation ended successfully")
        
    return pie_percentages1
    

def character_states_test(cladogram_dict, character_dict,
                          prefix, pdf_files=False):
    """
    Run the character states testing procedure for hierarchical characters
    defined by Cao (2008) and corrected by Rineau (2017). This procedure tests
    at each symmetric node (paralog) of the cladogram the presence of
    instances of the derived state.

          Cao, N. (2008). Analyse à trois éléments et anatomie du bois des
          Fagales Engl (Doctoral dissertation, Paris, Muséum national
          d'histoire naturelle).

          Rineau, V. (2017). Un nouveau regard cladistique sur l'anatomie
          comparée, la phylogénie, la systématique et la paléoécologie des
          rudistes (Bivalvia, Hippuritida) (Doctoral dissertation,
          Université Pierre et Marie Curie).

    Parameters
    ----------
    cladogram_dict : dict
        Dictionary containing one newick tree (ete3 Tree objects) as key.
        The tree must be the optimal cladogram obtained from the cladistic
        analysis of character trees stored in the character_dict argument.
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
        The trees must be the initial characters.
    prefix : str
        Prefix of the files to be computed.
    pdf_files : bool, optional
        If true, save a pdf file with one page for each character state for
        visualizing the results of the character state procedure.
        The default is False (no pdf files computed).

    Returns
    -------
    dict
        Two dictionaries with the results of the procedure sorted in two ways.

    """

    def find_synapomorphy(leaf_syn):
        """
        For one leaf, recursion over the tree in preorder and find the
        corresponding synapomorphy.

        """

        if leaf_syn.is_root():
            ancestor_syn = leaf_syn
        elif leaf_syn.get_ancestors()[0].state == "y":
            ancestor_syn = find_synapomorphy(leaf_syn.get_ancestors()[0])
        else:  # when acestor is find, captured here
            ancestor_syn = leaf_syn
        return ancestor_syn


    def synapomorphy_test_placement(cladogram, taxa_in_state,
                                    results_test_dict, char_names,
                                    char_states_names):
        """
            Test and placement of a character state.

        """
        results_test_dict[char_names][char_states_names] = []
        synapomorphies_set = set()  # set of nodes with derived state
        synapo_test = str("accepted")  #accepted(default)/rejected
        potential_synapo = cladogram.get_common_ancestor(taxa_in_state)

        # assignation of the state to a node
        # if no incongruence or no symmetric node implied
        if potential_synapo.get_leaf_names() == taxa_in_state or not (
                potential_synapo.search_nodes(clade_type="paralog")):
            synapomorphies_set.add(potential_synapo)

        # if ambiguous interpretation, check leaves in the symmetric nodes
        else:
            for node_cladogram in cladogram.traverse(strategy="postorder"):
                node_cladogram.add_feature("state", "n")  # no

            for leaf_node1 in cladogram.iter_leaves():
                if leaf_node1.name in taxa_in_state:
                    leaf_node1.state = "y"  # yes

            for node_cladogram in cladogram.traverse(strategy="postorder"):

                if node_cladogram.clade_type == "apical_node":
                    if len([apical_node_leaf for apical_node_leaf in
                            node_cladogram.children if
                            apical_node_leaf.state == "y"]) > 0:
                        node_cladogram.state = "y"

                elif node_cladogram.clade_type == "paralog":
                    test_node_cptr = 0
                    for test_node in (gen_paralog_node for gen_paralog_node in
                                      node_cladogram.children if
                                      gen_paralog_node.is_leaf() == False):
                        if test_node.state == "y":
                            test_node_cptr += 1
                    if test_node_cptr == len(node_cladogram.children):
                        node_cladogram.state = "y"

                elif node_cladogram.clade_type == "ortholog":
                    orth_leaf_gen = (child_orth_leaf for child_orth_leaf in
                                     node_cladogram.children if
                                     child_orth_leaf.is_leaf() == True)
                    orth_leaf_cptr = len([orth_leaf for orth_leaf in
                                          orth_leaf_gen if
                                          orth_leaf.state == "y"])
                    orth_paralog_gen = (node_child_orth for node_child_orth in
                                        node_cladogram.get_descendants() if
                                    node_child_orth.clade_type == "paralog")
                    test_n = len([paralog_node for paralog_node in
                                  orth_paralog_gen if
                                  paralog_node.state == "n"])
                    if test_n == 0 and orth_leaf_cptr > 0:
                        node_cladogram.state = "y"

            for orth_node in (node for node in cladogram.traverse(
                    strategy="postorder") if node.clade_type =="ortholog"
                    and node.state == "y"):  # fill ortholog lineages with y
                for orth_node_child in orth_node.get_descendants():
                    orth_node_child.state = "y"

            # gather nodes in synapomorphies_set
            for leaf_syn_y in (leaf_syn for leaf_syn in cladogram.iter_leaves()
                               if leaf_syn.state == "y"):
                synapomorphies_set.add(find_synapomorphy(leaf_syn_y))

        # check nb of retained nodes: 1 (syn, accepted), 2 or more = homoplasy
        if len(synapomorphies_set) > 1:
            synapo_test = str("rejected")
        results_test_dict[char_names][char_states_names] = [synapo_test,
                                                            synapomorphies_set]

        return synapo_test, synapomorphies_set, results_test_dict, cladogram

    pdfs = []

    if rep_detector(cladogram_dict):
        print("ERROR: Repeated leaves have been detected in the " +
                         "cladogram.\nOperation aborted.")
        sys.exit(1)
    elif rep_detector(character_dict):
        print("ERROR: Repeated leaves have been detected in the " +
                         "character tree set.\nOperation aborted.")
        sys.exit(1)
    else:
        for t, i in character_dict.items():
            for leaf in t.get_leaf_names():
                if not leaf in list(cladogram_dict.keys())[0].get_leaf_names():
                    print("ERROR: Character tree " + str(i)
                                   + ": leaf set must be equal"
                                   + " or a subset of the cladogram leaf "
                                   + "set.\nOperation aborted.")
                    sys.exit(1)

    # activate stylenodes from ete3 if pdf option
    if pdf_files:
        try:
            from ete3 import NodeStyle, TreeStyle, faces, TextFace

            # node styles(standard, synapomorphy, homoplasy)
            nstyle_other = NodeStyle()  # neutral style
            nstyle_other["shape"] = "circle"
            nstyle_other["size"] = 0.3
            nstyle_other["fgcolor"] = "black"

            nstyle_accept = NodeStyle()  # synapomorphy style
            nstyle_accept["shape"] = "sphere"
            nstyle_accept["size"] = 3
            nstyle_accept["fgcolor"] = "limegreen"
            nstyle_accept.opacity = 0.3

            nstyle_reject = NodeStyle()  # homoplasy style
            nstyle_reject["shape"] = "sphere"
            nstyle_reject["size"] = 3
            nstyle_reject["fgcolor"] = "crimson"
            nstyle_reject.opacity = 0.3

            # leaf styles added dynamically (layout attached to the tree style)
            def lstyle(node_style):
                """
                Assigns layouts to nodes.

                """

                # assign leaf style
                if node_style.is_leaf():
                    if node_style.taxa_in_state == "in":
                        name_face = TextFace(node_style.name,
                                             fgcolor="steelblue",
                                             fsize=2,
                                             bold=True)
                    elif node_style.taxa_in_state == "out":
                        name_face = TextFace(node_style.name,
                                             fgcolor="dimgray",
                                             fsize=2)
                    else:
                        name_face = TextFace(node_style.name + " (?)",
                                             fgcolor="lightgray",
                                             fsize=2)
                    name_face.margin_left = 2
                    faces.add_face_to_node(name_face,
                                           node_style,
                                           column=0,
                                           position='branch-right')

        except ImportError:  # issue with ete3 imports
            print("A manual instal of PyQt5 is requested to use the --pdf "
                  + "functionality\nPlease install using 'pip install" +
                  " PyQt5' and rerun the command line")
            sys.exit(1)

    print("Initiating character state test procedure")

    # each node is numbered and defined as paralog/ortholog/apical
    for cladogram in cladogram_dict:
        cladogram, syn_dict = cladogram_label(cladogram,
                                              clade_number_option="yes",
                                              clade_type_option="yes",
                                              pdf_files=pdf_files)

    # list character, states, and taxa content + character cardinal
    character_cardinal_dict = {}
    character_component_dict = {}
    
    # for each character
    for character, values in character_dict.items():  
        character_states = {}
        character_states_count = 0
        value = str(values).split('.')[0] # remove artificial number from poly.
        
        # for each character state
        for node in character.traverse(strategy="preorder"):
            if node.is_leaf() == False  and node.is_root() == False:
                
                checkstate = True
                
                #find charstate names from hmatrix, new numbers if not present
                try: 
                    character_states_count = node.charstate_name  # hmatrix
                except AttributeError:
                    if type(character_states_count) == int:
                        character_states_count += 1
                        
                    # case of states without names in hmatrix
                    elif type(character_states_count) == str:
                        childlist = []
                        for children in node.children:
                            try:
                                childlist.append(children.charstate_name)
                            except:
                                pass
                        
                        character_states_count = "Unnamed?"
                        character_states_count += '+'.join(childlist)
                        
                        character_states["Character state #" + str(value) + "." 
                     + str(character_states_count)] = Tree.get_leaf_names(node)
                        
                        character_cardinal_dict["Character state #" 
                     + str(value) + "." + str(character_states_count)] = set(
                         Tree.get_leaf_names(character))
                        
                        checkstate = False
                        
                    else:
                        print("ERROR: incorrect label for character states")
                        sys.exit(1)
                
                if checkstate:
                
                    character_states["Character state #" + str(value) + "." 
                     + str(character_states_count)] = Tree.get_leaf_names(node)
                    
                    character_cardinal_dict["Character state #" + str(value) 
                                             + "." + str(
                                             character_states_count)] = set(
                                             Tree.get_leaf_names(character))
                
        character_component_dict["Character_"+str(value)  + "." + str(
                                 character_states_count)] = character_states

    # character state test and adding node style
    results_test_dict = {}
    for cladogram in cladogram_dict:
        with open(prefix+".chartest", "w") as results_file:
            results_file.write("#Character states tests")
            results_file.write(("\n#Legend: #character.state / state accepted"
                               " or rejected / node(s) characterised by the"
                               " state (if one: synapomorphy)"))
            results_file.write("\n" + cladogram.write(format=8) + "\n")

        # for each character
        for char_names, values in character_component_dict.items():
            results_test_dict[char_names] = {}

            #for each character state
            for char_states_names, taxa_in_state in values.items():

                # character state test
                if character_cardinal_dict[char_states_names] == set(
                        Tree.get_leaf_names(cladogram)):  # if no missing data
                    synapomorphy_test, synapomorphies_set, results_test_dict, \
                    cladogram = synapomorphy_test_placement(
                        cladogram, taxa_in_state, results_test_dict,
                        char_names, char_states_names)

                else:  # if there is missing data
                    pruned_cladogram = cladogram.copy(method='cpickle')
                    pruned_cladogram.prune(list(
                        character_cardinal_dict[char_states_names]))  #prune
                    pruned_cladogram, temp = cladogram_label(pruned_cladogram,
                             clade_number_option="no", clade_type_option="yes",
                             pdf_files=pdf_files)
                    synapomorphy_test, synapomorphies_set, results_test_dict, \
                        pruned_cladogram = synapomorphy_test_placement(
                            pruned_cladogram, taxa_in_state, results_test_dict,
                            char_names, char_states_names)

                    syn_list_extension = {}  # dict filled with synapomorphies
                    for synapomorphy in synapomorphies_set:
                        if synapomorphy.is_leaf():
                            syn_list_extension[synapomorphy] = [
                                synapomorphy.name]
                        elif synapomorphy.clade_type == "apical_node":
                            syn_list_extension[synapomorphy] = [leaf.name for
                            leaf in synapomorphy.iter_leaves() if leaf.name in
                            taxa_in_state]
                        else:
                            syn_list_extension[synapomorphy] = [leaf.name for
                            leaf in synapomorphy.iter_leaves()] #syn/leaf_list

                    synapomorphies_set_2 = set()
                    synapomorphies_set = set()

                    #dissociate internal nodes and leaves
                    for syn_list in syn_list_extension.values():
                        if len(syn_list) == 1:  # if leaf
                            synapomorphies_set_2.add(
                                cladogram.get_leaves_by_name(syn_list[0])[0])
                        else:
                            synapomorphies_set.add(
                                cladogram.get_common_ancestor(
                                    [cladogram.get_leaves_by_name(leaf_name)[0]
                                     for leaf_name in syn_list]))

                    # correction of y-leaves branched to apical nodes
                    for node in synapomorphies_set_2:
                        if node.get_ancestors()[0].clade_type == "apical_node":
                            synapomorphies_set.add(node.get_ancestors()[0])
                        elif (node.get_ancestors()[0].clade_type ==
                              "ortholog") and set([node_name.name for node_name
                              in node.get_ancestors()[0].get_leaves()]) & set(
                                  taxa_in_state) == {node.name}:
                            synapomorphies_set.add(node.get_ancestors()[0])
                        else:
                            synapomorphies_set.add(node)

                    # synapomorphies_set updated with the non pruned cladogram
                    results_test_dict[char_names][char_states_names] = [
                        synapomorphy_test, synapomorphies_set]

                # add leaf style for pdf files
                if pdf_files:

                    synapomorphy_style = TreeStyle()
                    synapomorphy_style.show_leaf_name = True
                    synapomorphy_style.margin_top = 10
                    synapomorphy_style.margin_right = 10
                    synapomorphy_style.margin_left = 10
                    synapomorphy_style.margin_bottom = 10
                    synapomorphy_style.show_scale = False
                    synapomorphy_style.show_leaf_name = False
                    synapomorphy_style.title.add_face(TextFace(
                        char_states_names, fsize=3), column=0) # pdf title

                    for node1 in cladogram.iter_leaves():
                        if node1.name in taxa_in_state:
                            node1.add_feature("taxa_in_state", "in")
                        elif node1.name in character_cardinal_dict[
                                char_states_names]:
                            node1.add_feature("taxa_in_state", "out")
                        else:
                            node1.add_feature("taxa_in_state", "?")

                    synapomorphy_style.layout_fn = lstyle # layout modif nodes

                    #ajout des styles de noeuds internes
                    for node_style in cladogram.traverse(strategy="postorder"):
                        if node_style in synapomorphies_set:
                            if synapomorphy_test == "accepted":
                                node_style.set_style(nstyle_accept)
                            else:
                                node_style.set_style(nstyle_reject)
                        else:
                            node_style.set_style(nstyle_other)

                    # add comments to the pdf
                    synapomorphy_style.legend.add_face(TextFace(
                        "Synapomorphy test: "+synapomorphy_test, fsize=2),
                        column=0)
                    synapomorphy_style.legend.add_face(TextFace(
                        "", fsize=2), column=0)
                    synapomorphy_style.legend.add_face(TextFace(
                        "Node(s): ", fsize=2), column=0)
                    node_line_pdf = str()
                    for node_text in sorted([syn_node.clade_label for syn_node
                                             in synapomorphies_set
                                             if not syn_node.is_leaf()]):
                        node_line_pdf += str(node_text)
                        node_line_pdf += " "
                    synapomorphy_style.legend.add_face(TextFace(
                        node_line_pdf, fsize=2), column=0)

                    node_line_pdf = str()
                    for node_text in sorted([syn_node.clade_label for
                                             syn_node in synapomorphies_set if
                                             syn_node.is_leaf()]):
                        node_line_pdf += str(node_text)
                        node_line_pdf += " "
                    synapomorphy_style.legend.add_face(TextFace(
                        node_line_pdf, fsize=2), column=0)

                    # save pdf
                    cladogram.render(char_states_names + ".pdf",
                                     tree_style=synapomorphy_style)

                    pdfs.append(char_states_names + ".pdf")

                # append text file
                with open(prefix+".chartest", "a") as results_file:
                    results_file.write("\n" + char_states_names + ": "
                                       + synapomorphy_test)
                    syn_not_leaf = sorted([syn_node.clade_label for syn_node in
                                           synapomorphies_set if not
                                           syn_node.is_leaf()])
                    syn_leaf = sorted([syn_node.clade_label for syn_node in
                                       synapomorphies_set if
                                       syn_node.is_leaf()])
                    if len(syn_not_leaf) == 1:
                        results_file.write(", clade ")
                    elif len(syn_not_leaf) > 1:
                        results_file.write(", clades ")
                    for node_syn in syn_not_leaf:
                        results_file.write(str(node_syn)+", ")
                    for node_syn in syn_leaf:
                        results_file.write(node_syn+", ")

    # write file with synapomorphies location
    for char_names, values1 in results_test_dict.items():
        for char_states_names, syn_test_set in values1.items():
            for node_set in syn_test_set[1]:
                if syn_test_set[0] == "accepted":
                    syn_dict[str(node_set.clade_label)]["accepted"].append(
                        [char_states_names.replace("Character state ","")])
                elif syn_test_set[0] == "rejected":
                    syn_dict[str(node_set.clade_label)]["rejected"].append([
                        char_states_names.replace("Character state ","")])

    with open(prefix+".chartest_node", "w") as results_file_tree:
        results_file_tree.write("Synapomorphies by node:")
        results_file_tree.write("\n" + cladogram.write(format=8) + "\n")
        for node_set in sorted((str(node_set) for node_set, values in
                                syn_dict.items())):
            if syn_dict[node_set]["accepted"] or (
                    syn_dict[node_set]["rejected"]):
                results_file_tree.write("\n# "+node_set)
                results_file_tree.write("\n     Synapomorphies : ")
                if syn_dict[node_set]["accepted"]:
                    for node_syn in syn_dict[node_set]["accepted"]:
                        results_file_tree.write(str(node_syn)+", ")
                results_file_tree.write("\n     Homoplasies : ")
                if syn_dict[node_set]["rejected"]:
                    for node_h in syn_dict[node_set]["rejected"]:
                        results_file_tree.write(str(node_h)+", ")
                results_file_tree.write("\n")

    if pdf_files:  # merge all pdfs

        merger = PdfMerger()

        for pdf in pdfs:
            merger.append(pdf)

        merger.write(prefix + ".pdf")
        merger.close()

        for pdf in pdfs:
            remove(pdf)

    print("Character state test procedure ended successfully")

    return results_test_dict, syn_dict


def describe_forest(character_dict, prefix, showtaxanames=False):
    """
    Write on a text file basic informations about rooted trees: number of
    nodes, terminals, internal nodes, symmetric nodes, apical nodes.
    Resolution of the tree, number of dichotomies, polytomies.
    Optionaly, the list of leaves can be writen on the file.

    Parameters
    ----------
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.
    prefix : str
        Prefix of the output file.
    showtaxanames : bool, optional
        If true, write the list of leaves of the tree in the output file.
        The default is False.

    Returns
    -------
    None.

    """

    def describe_tree(dtree, outfile, nb=1, showtaxanames=False):
        """
        Write basic informations about a single rooted tree.

        Parameters
        ----------
        dtree : ete3.Tree
            A single rooted tree.
        nb : integer
            number of the tree (default 1).

        Returns
        -------
        None.

        """

        terminals = sorted(dtree.get_leaf_names())
        cardinal = len(terminals)
        paralogn = 0
        orthologn = 0
        apicaln = 0
        totaln = 0
        dichotomies = 0
        polytomies = 0
        polytomiesd = dict()

        taxalist = taxa_extraction({dtree:0})
        repetitions = len(taxalist) - len(set(taxalist))

        for node in dtree.traverse("postorder"):
            if not node.is_leaf():  # if internal node

                totaln += 1
                i = len([child_node for child_node in node.get_children(
                    ) if not child_node.is_leaf()])
                j = len([child_node for child_node in node.get_children()])

                if i >= 2:  # paralog
                    paralogn += 1
                elif i == 1:  # ortholog
                    orthologn += 1
                elif i == 0:  # apical
                    apicaln += 1

                if j == 2:  # dichotomies vs polytomies
                    dichotomies += 1
                else:
                    polytomies += 1

                    if j in polytomiesd: # details about polytomies
                        polytomiesd[j] += 1
                    else:
                        polytomiesd[j] = 1

        with open(outfile, "a") as describefile:
            describefile.write("***************\n")
            describefile.write("Describe tree {}\n".format(str(nb)))
            describefile.write("Number of nodes: " + str(totaln + cardinal)
                               + "\n")
            describefile.write("    Terminals: " + str(cardinal)
                               + "\n")
            describefile.write("    Internal nodes: " + str(totaln)
                               + "\n")
            describefile.write("        Symmetric nodes: " + str(paralogn)
                               + "\n")
            describefile.write("        Asymmetric nodes: " + str(orthologn)
                               + "\n")
            describefile.write("        Apical nodes: " + str(apicaln)
                               + "\n")
            describefile.write("\n")

            if cardinal > 2:
                describefile.write("Resolution: " + str(
                    round(((totaln - 1) / (cardinal - 2)) * 100, 5)) + "%\n")
            else:
                describefile.write("Resolution: NA\n")

            describefile.write("Dichotomies: {}; polytomies: {}\n".format(
                str(dichotomies),str(polytomies)))
            describefile.write("Repetitions: {}\n".format(str(repetitions)))

            if polytomies > 0:
                describefile.write("Polytomies - details\n" +
                                   "(level: number of occurences)\n")
                for poly in sorted(polytomiesd.keys()):
                    describefile.write("{} : {}\n".format(str(poly),
                                                    str(polytomiesd[poly])))

            if showtaxanames:
                describefile.write("\nTaxa list:\n")
                for taxa in taxalist:
                    describefile.write(taxa+"\n")

            describefile.write("\n")

    i = 0
    for dtree in character_dict.keys():
        i += 1
        describe_tree(dtree, prefix+".dt", nb=i, showtaxanames=showtaxanames)

    print("Forest described")


def chartest(cladopath, charpaths, taxarep1=False, taxarep2=False, method="TMS", 
             prefix="AGATTA_chartest", pdf_files=False, verbose=False):
    """
    Run the character states testing procedure for hierarchical characters
    defined by Cao (2008) and corrected by Rineau (2017). This procedure tests
    at each symmetric node (paralog) of the cladogram the presence of
    instances of the derived state.

          Cao, N. (2008). Analyse à trois éléments et anatomie du bois des
          Fagales Engl (Doctoral dissertation, Paris, Muséum national
          d'histoire naturelle).

          Rineau, V. (2017). Un nouveau regard cladistique sur l'anatomie
          comparée, la phylogénie, la systématique et la paléoécologie des
          rudistes (Bivalvia, Hippuritida) (Doctoral dissertation,
          Université Pierre et Marie Curie).

    This function is made for giving trees directly to character_states_test
    from newick files

    Parameters
    ----------
    cladopath : str
        Path to a newick file containing the cladogram. The tree must be the
        optimal cladogram obtained from the cladistic analysis of character
        trees stored in charpaths.
    charpaths : list of str
        Path(s) to file(s) containing the initial character trees in newick or 
        hmatrix format.
    taxarep1 : str, optional
        DESCRIPTION. The default is False.
    taxarep2 : str, optional
        DESCRIPTION. The default is False.
    method : str, optional
        One of the two implemented algorithms of free-paralogy subtree
        analysis between "TMS" and "FPS" for removing polymorphism. 
        The default is "TMS".
    prefix : str
        Prefix of the files to be computed.
    pdf_files : str or bool, optional
        This argument can be the path stating the location where to save one
        pdf file for each character state for visualizing the results of the
        character state procedure.
        The default is False (no pdf files computed).
    verbose : bool, optional
        Verbose mode if True. The default is True.
        
    Returns
    -------
    None.

    """

    print("Loading cladogram")

    cladopath = path.expanduser(cladopath)
    charpaths = [path.expanduser(charpath) for charpath in charpaths]

    cladogram_dict = character_extraction(cladopath, taxarep1, verbose=False)

    print("Cladogram loaded")

    character_dict = wrapper_character_extraction(charpaths, 
                                                  taxarep2,
                                                  prefix,
                                                  verbose)
        
    # remove automatically repetitions if detected (user message printed)
    if rep_detector(character_dict):
        character_dict = del_replications_forest(character_dict,
                                                 method=method,
                                                 prefix=prefix,
                                                 verbose=verbose)

    # print(str(len(character_dict)) + " characters loaded")

    character_states_test(cladogram_dict, character_dict, prefix, pdf_files)
    
def brlen(prefix, cladogram_dict_list):
    """
    Compute node support for each resulting cladogram based 
    on triplets (amount weighted triplets compatible with each node). 

    Parameters
    ----------
    prefix : TYPE
        Prefix of the triplet file.
    cladogram_dict_list : dict or list
        dictionary or list of cladograms computed by the analysis.

    """
    
    print("Node support computation")
    
    triplet_dict = triplet_extraction(prefix+'.triplet', 
                                taxa_replacement_file=prefix+'.taxabloc')
    
    if type(cladogram_dict_list) == dict:
        for tree in cladogram_dict_list.keys():
            for node in tree.traverse():
                if not node.is_leaf() and not node.is_root():
                    
                    node.dist = 0  # set support value to 0
                    
                    # for each node, compute support values
                    for trip, FW in triplet_dict.items():
                        
                        in1, in2 = trip.in_taxa    
                        out, = trip.out_taxa    
                        
                        if (in1 in node.get_leaf_names() 
                            and in2 in node.get_leaf_names() 
                            and out not in node.get_leaf_names()):
                            
                            node.dist += FW
    
    elif type(cladogram_dict_list) == list:
        for tree in cladogram_dict_list:
            for node in tree.traverse():
                if not node.is_leaf() and not node.is_root():
                    
                    node.dist = 0  # set support value to 0
                    
                    # for each node, compute support values
                    for trip, FW in triplet_dict.items():
                        
                        in1, in2 = trip.in_taxa    
                        out, = trip.out_taxa    
                        
                        if (in1 in node.get_leaf_names() 
                            and in2 in node.get_leaf_names() 
                            and out not in node.get_leaf_names()):
                            
                            node.dist += FW
                                    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

    Agatta: Three-item analysis Python package
    Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

    Agatta is a set of tools in the cladistic framework to perform
    three-item analysis and associated operations in cladistics in python.

    https://github.com/vrineau/agatta

    This code is under license GNU GPLv3

"""

__version__ = "0.9.7"



# -*- coding: utf-8 -*-
"""

    Agatta: Three-item analysis Python package
    Contact: Valentin Rineau - valentin.rineau@sorbonne-universite.fr

    Agatta is a set of tools in the cladistic framework to perform
    three-item analysis and associated operations in cladistics in python.

    https://github.com/vrineau/agatta

    This code is under license GNU GPLv3

"""

from os import path
from re import search
from re import findall
import csv
import sys
import platform
import treeswift
from ete3 import Tree
from ete3.parser.newick import NewickError


def character_extraction(infile=False, taxa_replacement=False, verbose=True,
                         info_tree=True):
    """
    Extract newick rooted trees from a text file.
    The function extract lines begining with an open parenthesis and
    closing with a semicolon and interpret them as newick strings.
    The first argument is the path of the file, but the function can be used
    without any argument. In this case, a file selection window appears.
    A taxa bloc can also be present to replace taxa names.

    Parameters
    ----------
    infile : str
        Path of the file containing rooted trees in newick format
        (https://en.wikipedia.org/wiki/Newick_format). The input file must
        contain a single newick string on each line. Whitespaces are not
        allowed for leaf names.
        Example:

            (a,b,(c,d,(e,f)));
            ((c,e),(a,(b,(d,f))));

    taxa_replacement : bool, optional
        Path of a table file containing two columns. The first column
        corresponds to the names of the terminals of the newick stored in
        infile, and the second column corresponds to their names the user wants
        to obtain at the end. All separators accepted. Example:
             AA Diceras
             AB Valletia
             AC Monopleura
        The default is False (no replacement).
    info_tree : bool, optional
        Option that discard non-informative trees (at least, two terminals that 
        are closer together relatively to a third one) when activated.
        The default is True (non-informative trees are discarded).

    Returns
    -------
    character_dict : dict
        Dictionnary containing trees in keys with indices in values.

    """

    if verbose:
        print("Loading character trees")

    character_dict = {}  # return dictionary containing trees
    taxa_dict = {}  # taxa / symbol dictionary used if taxa_replacement == True

    # selection file window
    # if infile is None:
    #     root = Tk()
    #     infile = filedialog.askopenfilename(
    #         title="Select file containing newick character trees",
    #         filetypes=(
    #             ("Newick files", "*.*"),
    #             ("Lisbeth files", "*.3ia"),
    #             ("Nexus files", "*.nex"),
    #             ("Nexml files", "*.nexml")))
    #     root.withdraw()

    infile = path.expanduser(infile)

    if not path.isfile(infile):
        print("ERROR: The file '" + infile + "' does not exist." +
                       "\nOperation aborted.")
        sys.exit(1)

    fileName, fileExtension = path.splitext(infile)
    a = 1

    lisbethchecker = True

    if fileExtension == ".3ia":
        brokennewick = ""
        with open(infile, "r") as file_tree :
            line_nb = 0
            for line in file_tree:
                line_nb += 1
                for character_newick in findall(r"\(\S+", line):
                    if character_newick:
                        is_exception = True
                        try:
                            character_dict[Tree(character_newick + ';')] = a
                            a += 1
                        except NewickError:
                            brokennewick += ("ERROR: " +
                                "Line " + str(line_nb) +
                                ": Broken newick structure")
                        else:
                            no_exception = True
                        if not 'no_exception' in locals():
                            sys.exit(1)

                if search(r"\s=\s", line):
                    taxa_dict[line.split(" = ")[0].split()[0]] = line.split(
                        " = ")[1].strip()
                    lisbethchecker = False

        if lisbethchecker:  # if not 3ia file
            print("WARNING: the file doesn't seems to be a .3ia file." +
                  " Please check the file extension.")

        print(brokennewick)

    else:
        try:
            if fileExtension == ".nex":
                tstreelist = treeswift.read_tree_nexus(infile)

            elif fileExtension == ".nexml":
                tstreelist = treeswift.read_tree_nexml(infile)

            else:  # newick files
                tstreelist = treeswift.read_tree_newick(infile)
        except:
            print("ERROR: The file " + infile + " is broken. Please check " +
                           "the format. \nNexus files must have the .nex " +
                           "extension.\nNexml files must have the .nexml " +
                           "extension.\nHierarchical matrices have" +
                           " the .hmatrix extension.\nAll other extensions " +
                           "are considered as newick files containing only " +
                           "newick strings (one on each line).")
        else:
            no_exception = True
        if not 'no_exception' in locals():
            sys.exit(1)

        if not isinstance(tstreelist, list):
            tstreelist = [tstreelist]

        for tstree in tstreelist:
            if isinstance(tstree, dict):
                for idtree, tst in tstree.items():
                    try:
                        if (tst.newick().startswith("[&R] ") or
                            tst.newick().startswith("[&U] ")):
                            character_dict[Tree(
                                tst.newick().split(" ")[1])] = idtree
                        else:
                            character_dict[Tree(tst.newick())] = idtree
                    except NewickError:
                        print("Tree {}: Broken newick structure.".format(
                            str(idtree)))
                    else:
                        no_exception = True
                    if not 'no_exception' in locals():
                        sys.exit(1)

            else:
                try:
                    if (tstree.newick().startswith("[&R] ") or
                        tstree.newick().startswith("[&U] ")):
                        character_dict[Tree(tstree.newick().split(" ")[1])] = a
                    else:
                        character_dict[Tree(tstree.newick())] = a
                    a += 1
                except NewickError:
                    print("Tree {}: Broken newick structure.".format(str(a)))
                else:
                    no_exception = True
                if not 'no_exception' in locals():
                    sys.exit(1)

    if taxa_replacement:
        if not path.isfile(taxa_replacement):
            print("ERROR: The input file '" + taxa_replacement
                                    + "' does not exist."
                                    + "\nOperation aborted.")
            sys.exit(1)

        with open(taxa_replacement, "r") as taxa_table:
            try:
                dialect = csv.Sniffer().sniff(taxa_table.readline(),
                                              delimiters=[";","\t"," ","|"])
            except:
                print("ERROR: The table file '" + taxa_replacement
                               + "' is not adequately formated."
                               + "\n Operation aborted.")
            else:
                no_exception = True
            if not 'no_exception' in locals():
                sys.exit(1)

        with open(taxa_replacement, "r") as taxa_table:
            tab_test = list(csv.reader(taxa_table,
                                       delimiter=dialect.delimiter))

            for rowlist in tab_test:  # remove trailing spaces
                rowlist = [e.strip() for e in rowlist]

            if not len({len(l) for l in tab_test}) == 1:
                 print("ERROR: The table file '"
                                   + taxa_replacement + "' is broken."
                                   + "\nOperation aborted.")
                 sys.exit(1)

            for idtax, nametax in tab_test:
                taxa_dict[idtax] = nametax

    if taxa_replacement or fileExtension == ".3ia":
        for cladogram in character_dict.keys():
            for leaf in cladogram.iter_leaves():
                try:
                    leaf.name = taxa_dict[leaf.name]
                except KeyError:
                    if not lisbethchecker:
                        print("ERROR: The name '" + str(leaf.name)
                                       + "' cannot be replaced."
                                       + "\nOperation aborted.")
                else:
                    no_exception = True
                if not 'no_exception' in locals():
                    sys.exit(1)

    if verbose:
        print("{} character(s) loaded".format(str(len(character_dict))))

    # non-informative trees are discarded
    if info_tree:
        character_dict, noninfo = infotree_checker(character_dict)
        
    # empty nodes automatically removed
    delnodes = []  # list of nodes to delete
    for char, nb in character_dict.items():
        error_empty = False
        for node in char.traverse():
            for childnode in node.get_children():
                if (sorted(node.get_leaf_names()) == sorted(
                        childnode.get_leaf_names())):
                    delnodes.append(childnode)  #delete node
                    error_empty = True
        if error_empty:
            print("WARNING: empty node(s) in character " + str(nb)
                  + " detected and automatically removed.")
                    
    # deletion of state branches
    for delnode in delnodes:
        delnode.delete()

    return character_dict


def infotree_checker(character_dict, verbose=True):
    """
       Check if all characters are informative and returns a dictionary
       without non-informative trees
    """

    non_info_chars = []
    info_character_dict = dict()
    non_info_characters = list()

    for cladogram, a in character_dict.items():
        info = False
        lroot = len(cladogram.get_leaf_names())
        non_info_chars.append(a)
        if lroot > 2:
            for node in cladogram.traverse():
                if not node.is_leaf() and not node.is_root():
                    lnode = len(node.get_leaf_names())
                    if lnode > 1 and lnode < lroot:  # if tree is informative
                        non_info_chars.remove(a)
                        info_character_dict[cladogram] = a
                        info = True
                        break
        if not info:
            non_info_characters.append(a)

    if non_info_chars and verbose:
        print("WARNING: the following input trees are non-informative:")
        for a in non_info_chars:
            non_info_tree = list(character_dict.keys())[list(
                character_dict.values()).index(a)].write(format=9)
            print("[{}] {}".format(str(a),non_info_tree))

    return info_character_dict, non_info_characters


def taxa_extraction(character_dict):
    """
    Extracts all leaves from a dictionary containing newick trees in keys.

    Parameters
    ----------
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.

    Returns
    -------
    taxa_set : set
        set of taxa names.

    """

    taxa_set = set()
    for character in character_dict.keys():
        taxa_set.update(set(Tree.get_leaf_names(character)))
    return taxa_set


def taxa_triplet_extraction(triplet_dict):
    """
    Extracts all leaves from a dictionary containing triplets in keys.

    Parameters
    ----------
    triplet_dict : dict
        Dictionary containing triplets (Agatta triplet objects) in keys.

    Returns
    -------
    taxa_set : set
        set of taxa names.

    """

    taxa_set = set()

    for triplet in triplet_dict.keys():
        taxa_set.update(triplet.out_taxa)
        taxa_set.update(triplet.in_taxa)

    return taxa_set


def taxa_to_numbers(character_dict):
    """
    Extracts all leaves from a dictionary containing newick trees in keys,
    and assign to each taxa name an identifier.

    Parameters
    ----------
    character_dict : dict
        Dictionary containing newick trees (ete3 Tree objects) in keys.

    Returns
    -------
    taxa_dict : dict
        Dictionary with taxa names as keys and identifiers as values.
    code_dict : dict
        Dictionary inverse of taxa_dict used for translation of identifiers.
    taxa_convers : list
        list of strings with taxa names and corresponding id.

    """
    taxa_set = taxa_extraction(character_dict)
    taxa_dict = {}
    code_dict = {}
    taxa_convers = []
    taxa_cptr = 1

    for taxa1 in taxa_set:
        taxa_dict[taxa1] = taxa_cptr
        code_dict[taxa_cptr] = taxa1
        taxa_convers.append(str(taxa1)+" "+str(taxa_cptr))
        taxa_cptr += 1

    return taxa_dict, code_dict, taxa_convers


def standardisation(tree_file, biogeo_tab, prefix, verbose=False):
    """
    This function is used for standardisation of characters in the context of
    cladistic biogeography, i.e., replacement of leaves using a correspondence
    table. In cladistic theory, the standardisation is the construction of
    character trees (hypotheses of kinship relationship between bearers) based
    on homologies (hypotheses of kinship relationships between parts). The
    standardisation is used to convert phylogenies in areagrams in cladistic
    biogeography.
    MAST are automatically removed.
    A file named prefix.stand is generated at the end.

    Parameters
    ----------
    tree_file : list of str or dict
        Path(s) of the file(s) containing a single newick tree on each line.
        Example:

            (a,b,(c,d,(e,f)));
            ((c,e),(a,(b,(d,f))));
        
        Can be also a character_dict
        The default is False (a selection window appears in this case).
    biogeo_tab : str
        path of table with semicolon separators and two columns, taxa in
        the left column and corresponding areas in the right column.
    prefix : str, optional
        prefix of the .stand file to save.
    verbose : bool, optional
        If true, print in the terminal the resulting areagrams and informations
        on taxa, areas, MASTs (Multiple Area Single Taxa) and repetitions.
        The default is False.

    Returns
    -------
    areagram_dict : dict
        Dictionary with areagram trees in keys (ete Tree objects) and same
        index as character_dict to track the original trees.

    """

    def biogeo_table(biogeo_tab, verbose=False):
        """
        This function requires in input a table (semicolon separators) with
        the correspondence between parts in the left column (e.g., taxa in
        cladistic biogeography) and bearers in the right column (e.g., areas)

        Parameters
        ----------
        biogeo_tab : str
            path of table file with two columns, taxa in
            the left column and corresponding areas in the right column.
            Example:

                Clinocaprina USA
                Titanosarcolites Jamaica
                Diceras France

        verbose : bool, optional
            If true, print in the terminal informations on taxa, areas, MASTs
            (Multiple Area Single Taxa) and repetitions. The default is False.

        Returns
        -------
        biogeo_dict : dict
            dictionary with taxa in keys and areas in values.
        MAST : set
            set of MAST areas.
        arearep : set
            set of repeated areas.

        """

        biogeo_dict = {}
        taxa = set()    # taxa set
        areas = set()   # areas set
        MAST = set()    # set of detected MAST
        arearep = set()  # set repeated leaves

        biogeo_tab = path.expanduser(biogeo_tab)

        with open(biogeo_tab, "r") as bt_file:
            try:
                dialect = csv.Sniffer().sniff(bt_file.readline(),
                                              delimiters=[";","\t"," ","|"])
            except:
                print("ERROR: The table file '" + biogeo_tab +
                                  "' is not adequately formated." +
                                  "\n Operation aborted.")
            else:
                no_exception = True
            if not 'no_exception' in locals():
                sys.exit(1)

        with open(biogeo_tab, "r") as bt_file:

            tab_test = csv.reader(bt_file, delimiter=dialect.delimiter)
            table = list(tab_test)

            for rowlist in table:  # remove trailing spaces
                rowlist = [e.strip() for e in rowlist]

            if not len({len(l) for l in table}) == 1:
                 print("ERROR: The table file '" + biogeo_tab +
                                  "' is broken.\n Operation aborted.")
                 sys.exit(1)

        for line in table:
            # detection of MASTs (if taxon already in list)
            if line[0] in taxa:  # if taxa already recorded = MAST
                biogeo_dict[line[0]].append(line[1])
                MAST.add(line[0])

            else:
                biogeo_dict[line[0]] = [line[1]]

            # detection of repeated areas
            if line[1] in areas:
                arearep.add(line[1])

            taxa.add(line[0])
            areas.add(line[1])

        if verbose:
            print(str(len(taxa)) + " terminal taxa, " + str(len(areas)) +
                  " terminal areas")
            print(str(len(arearep)) + " repeated taxa, " + str(len(MAST)) +
                  " MAST detected")
            print("taxa: {}".format(taxa))
            print("areas: {}".format(areas))
            if len(arearep) > 0:
                print("repeated taxa: {}".format(arearep))
            if len(MAST) > 0:
                print("MAST: {}".format(MAST))

        return biogeo_dict, MAST, arearep

    biogeo_dict, MAST, arearep = biogeo_table(biogeo_tab, verbose=False)
    
    if type(tree_file) == list:
        taxa_replacement=False
        chardec=False
        info_tree=True
        character_dict = wrapper_character_extraction(tree_file, 
                                                      taxa_replacement, prefix, 
                                                      chardec, verbose, 
                                                      info_tree)
    else:
        character_dict = tree_file
        
    areagram_dict = dict()
    nb_char = len(character_dict)

    print("Character standardisation")
    
    error_message = ""

    # for each tree
    with open(prefix + ".stand", "w") as area_file:
        for phylogeny, index in character_dict.items():

            phylogeny2 = phylogeny.copy()

            # MAST deletion
            for leaf in phylogeny.get_leaves():
                if leaf.name in MAST:
                    leaveswmast = list(set(phylogeny2.get_leaf_names()) - MAST)
                    phylogeny2.prune(leaveswmast)

            # for each terminal taxa, replacement by its corresponding area
            # MAST excepted
            for leaf in phylogeny.get_leaves():
                if leaf.name not in MAST:
                    for leafnode in phylogeny2.get_leaves_by_name(leaf.name):
                        try:
                            leafnode.name = biogeo_dict[leaf.name][0]
                        except KeyError:
                            error_message += ("ERROR: The name '"
                                            + str(leaf.name)
                                            + "' does not exists in the table"
                                            + " file '" + biogeo_tab + "'.\n")

            # add new areagram
            areagram_dict[phylogeny2] = index
            new_tree_line = "[" + str(index) + "] "
            new_tree_line += phylogeny.write(format=9)
            new_tree_line += "\n    "
            new_tree_line += phylogeny2.write(format=9)
            new_tree_line += "\n\n"

            area_file.write(new_tree_line)

            if verbose:
                print(new_tree_line)

    if error_message:
        print(error_message)
        print("Operation aborted.")
        sys.exit(1)

    print("Multiple Area Single Taxa (MAST) removed")
    print("Characters standardised\n")
    
    areagram_dict, noninfo = infotree_checker(areagram_dict)
    
    # message when non-informative characters
    for n in noninfo:
        print("Character {} non-informative after standardisation".format(n))

    return areagram_dict


def hmatrix(infile, prefix=False, chardec=False, verbose=False):
    """
    Function that build ete3 trees from a file containing a hierarchical
    matrix. The matrix must be a table text
    (e.g., a csv file). The separator cannot be a comma.

    Format of the matrix:

      * The first line corresponds to the hierarchical structure of the
        character tree with labels.
        It is a newick tree without ending semicolon and without leaves.
        It is to the user to choose which nodes are labelled. Labels must be
        integers.

        Examples: (0(1)), (0(1)(2(3(4)))), ((0)(1)).

      * The first column corresponds to taxa names. Semicolons cannot be used
        in a name. The first cell is empty as it corresponds to the line
        of hierarchical characters.

      * The following lines are identical to those of a standard
        taxon/character matrix. Each state corresponds to the branching
        position of a specific taxon on the character.
        Each cell can be filled by:

            - A question mark "?" or an empty cell if the state for a taxon is 
              unknown.

            - An integer which should be mandatory present in the first line
              of the matrix. For example, if the character is coded as (0,(1))
              by the user, only 0 and 1 can be used.

            - In case of polymorphism (if the user want to branch several
              instances of the same taxon in several nodes), the several
              integers must be separated by a comma ",".

        Example:

            ;(0,(1));(0,(1));(0,(1,(2)));(0,(1,(2),(3)))
            Diceras;0;0;0;0
            Hippurites;0;0;0;1
            Radiolites;1;0;2;2
            Titanosarcolites;1;1;2;2,3
            Clinocaprina;1;1;1;3
            Vaccinites;0,1;1;2;3


    Parameters
    ----------
    infile : str
        Path of file containing the hierarchical matrix.
    prefix : str, optional
        Prefix of the file to save. If prefix is set to false, no file is
        saved. The default is False.
    chardec : bool, optional
        If true, decompose each tree into components (one tree for each
        informative node). The default is False.

    Returns
    -------
    character_dict : dict
        Dictionnary containing trees in keys with indices in values.

    """

    def extracthmatrix(infile):
        """
        Extracts a hierarchical matrix from a csv file
        """

        if not path.isfile(infile):
            print("ERROR: The hierarchical matrix '" + infile
                                    + "' does not exist.\nOperation aborted.")
            sys.exit(1)

        with open(infile, 'rt') as f:
            try:
                dialect = csv.Sniffer().sniff(f.readline(),
                                              delimiters=[";","\t","|"])
            except:
                print("ERROR: The table file '" + infile +
                      "' is not adequately formated.\n Operation aborted.")
            else:
                no_exception = True
            if not 'no_exception' in locals():
                sys.exit(1)

        with open(infile, 'rt') as f:

            data = csv.reader(f, delimiter=dialect.delimiter)
            hmatrix = list(data)

            for rowlist in hmatrix:  # remove trailing spaces
                rowlist = [e.strip() for e in rowlist]

            if not len({len(l) for l in hmatrix}) == 1:
                 print("ERROR: the hierarchical matrix '" +
                                infile + "' is broken.\nOperation aborted.")
                 sys.exit(1)
                 
        if path.splitext(infile)[1] != ".hmatrix":
            print("WARNING: the matrix should have a .hmatrix extension.")

        return hmatrix

    if verbose:
        print("Loading hierarchical matrix")

    character_dict = dict()  # trees without polytomies
    temp_character_dict = dict()  # trees with polytomies (raw data)
    error_message = ""

    # read first matrix
    hmatrix = extracthmatrix(infile)

    # construction of character trees backbone
    i = 1
    treeliststr = hmatrix[0]
    del treeliststr[0]

    if [i for i in ''.join(treeliststr) if not (i in ["(",")",",","0","1",
                                                      "2","3","4","5","6",
                                                      "7","8","9"])]:
        print("ERROR: first line of the matrix is incorrect or missing")
        sys.exit(1)

    for char in treeliststr:
        try:
            temp_character_dict[Tree(char+";")] = str(i)
        except NewickError:
            print("ERROR: The newick tree " + char + " in column " +
                               str(i+1) + " is broken\nOperation aborted.")
        else:
            no_exception = True
        if not 'no_exception' in locals():
            sys.exit(1)
        i += 1

    taxalist = [hmatrix[ncol][0] for ncol in range(1, len(hmatrix))]

    # option for character decomposition into components
    if chardec:
        temp_character_dict_binary = dict()

        for char, ind in temp_character_dict.items():  # for each character

            # for each unrooted character state
            for state in char.traverse(strategy="preorder"):
                if state.is_leaf() and not state.get_ancestors()[0].is_root():

                    statetree = char.copy()

                    for isolstate in statetree.traverse(strategy="preorder"):
                        if not isolstate.is_leaf() and not isolstate.is_root():
                            if not isolstate.get_children()[0].name == \
                                                                    state.name:
                                isolstate.delete()

                    stateline = str(ind) + "." + str(state)
                    temp_character_dict_binary[statetree] = stateline

        temp_character_dict = temp_character_dict_binary

    # fill characters
    for char, ind in temp_character_dict.items():  # for each character
        # mark branches to delete at the end
        delnodes = []  # list of leaves to delete
        statelist = []
        for leaf in char.iter_leaves():
            statelist.append(leaf.name)
            delnodes.append(leaf)
            leaf.get_ancestors()[0].add_feature("charstate_name", leaf.name)
            leaf.name = "_agatta_charstate_"+leaf.name  # for avoiding the
            # specific case where character state and taxa have the same name
            
        # for each taxon
        for taxa in taxalist:
            if chardec:
                ind2 = ind.split(".")[0]
            else:
                ind2 = ind
            charstates = hmatrix[taxalist.index(taxa)+1][int(ind2)].split(",")

            if charstates[0] and charstates[0] != "?":  # if not missing data

                # branching
                for charstate in charstates:

                    # if incorrect matrix, raise error, otherwise, connect taxa
                    try:
                        branchnode = char.get_leaves_by_name(
                            "_agatta_charstate_" +
                            charstate)[0].get_ancestors()[0]
                        branchnode.add_child(name=taxa)
                    except:
                        error_message += "ERROR: The instance '"
                        error_message += str(charstate) + "' of leaf " + taxa
                        error_message += " in column " + str(ind2)
                        error_message += " does not match\n"

        # deletion of state branches
        for delnode in delnodes:
            delnode.delete()
            
        # handling ete3 bug: deletion of states with two empty parent nodes
        for leaf in char.traverse():
            if leaf.is_leaf():
                if not leaf.name:
                    leaf.delete()
        
        # autapomorphic (non-informative) state detection and warning
        for state in char.traverse():
            if not state.is_leaf() and not state.is_root():
                lnode = len(state.get_leaf_names())
                if lnode < 2:
                    print("WARNING: character " + str(ind2) 
                          + ", state " + str(state.charstate_name) 
                          + " non-informative")

    for char, value in temp_character_dict.items():
        if chardec:
            value2 = value.split(".")[0]
        else:
            value2 = value
        char.ladderize()
        character_dict[char] = str(value2)
        
    if error_message:
        print(error_message)
        sys.exit(1)

    character_dict, noninfo = infotree_checker(character_dict)

    # save resulting tree file
    if prefix:
        character_dict, noninfo = infotree_checker(character_dict)
        with open(prefix+".tre", "w") as treefile:
            for char, charnum in character_dict.items():
                treefile.write(char.write(format=9) + "\n")

    print("{} characters computed from the matrix".format(
                                                    str(len(character_dict))))

    return character_dict


def hmatrix_several(infile, prefix=False, chardec=False, verbose=False):
    """
    Wrapper of hmatrix for treating several matrices. Use only for hmatrix 
    Agatta command.

    Parameters
    ----------
    infile : list
        List of paths to hmatrix.
    prefix : str, optional
        Prefix of the file to save. If prefix is set to false, no file is
        saved. The default is False.
    chardec : bool, optional
        If true, decompose each tree into components (one tree for each
        informative node). The default is False.
    verbose : bool, optional
        Verbose mode. The default is False.

    Returns
    -------
    None.

    """
    
    for hfile in infile:
        hmatrix(hfile, prefix, chardec, verbose)

            
def wrapper_character_extraction(infilelist=False, taxa_replacement=False, 
                                 prefix=False, chardec=False, verbose=True, 
                                 info_tree=True):
    """
    extrait hmatrix ou newick et renvoie l'ensemble combiné

    Parameters
    ----------
    infilelist : list
        List of paths of files containing hierarchical matrices.

    taxa_replacement : TYPE, optional
        DESCRIPTION. The default is False.
    verbose : TYPE, optional
        DESCRIPTION. The default is True.
    info_tree : TYPE, optional
        DESCRIPTION. The default is True.

    Returns
    -------
    None.

    """
    
    cptr_file_nb = 0
    character_dict = dict()
    maxnb = 0
    
    # detect if one or several files
    if not isinstance(infilelist, list):
        infilelist = [infilelist]
    
    #for each file
    for infile in infilelist:
        
        cptr_file_nb +=1
        fileName, fileExtension = path.splitext(infile)

        if fileExtension == ".hmatrix":
            # if hmatrix file
            sub_chardict = hmatrix(infile, prefix, chardec, verbose)
                        
            print("File " + str(cptr_file_nb) + ": hierarchical matrix. "
                  + str(len(sub_chardict)) + " character(s) loaded")
            
        else:
            # if newick file
            sub_chardict = character_extraction(infile, taxa_replacement, 
                                 verbose, info_tree)
                        
            print("File {}: newick file. {} character(s) loaded".format(
                str(cptr_file_nb), str(len(sub_chardict))))
            
        # merge dictionaries
        for key, value in sub_chardict.items():
            if str(value).isdigit():
                character_dict[key] = str(int(value) + maxnb)
            else:  # si le nom du caractère n'est pas un numéro
                character_dict[key] = value
        
        # calculates the maximum number from which to continue numbering
        # if the character names are their numbers
        if all([str(x).isdigit() for x in character_dict.values()]):
            maxnb = max([int(x) for x in character_dict.values()])

           
    # count number of characters and where does they come from
    print("\nTreefication of {} characters complete".format(
        str(len(character_dict))))

    return character_dict
    

def checkargs(arguments):
    """
    Check parsing and exit if error in flags

    Parameters
    ----------
    arguments : dict
        Docopt dict.

    Returns
    -------
    None.

    """
    if arguments["analysis"]:

        if arguments["--pdf"] and not arguments["--chartest"]:
            sys.exit(print("ERROR: --pdf flag cannot be used without " +
                           "--chartest"))

        if (arguments["--software"] == "paup" or
            arguments["--software"] == "tnt" or
            arguments["--software"] == "wqfm" or
            arguments["--software"] == "wtree-qmc"):
            if (not arguments.get("--softpath", False)
                and platform.system() != "Windows"):
                sys.exit(print("ERROR: the path of the " +
                      "software is missing. Please use --softpath=file/path"))

        elif arguments["--software"] != "agatta":
            sys.exit(print("ERROR: --software flag must be one of: 'paup', " +
                           "'tnt', 'wqfm', 'wtree-qmc'"))

        if (arguments["--software"] == "wqfm"
            and not arguments["--analysis"] == "heuristic"):
            sys.exit(print("ERROR: WQFM works only in heuristic mode"))

        if (arguments["--software"] == "wtree-qmc"
            and not arguments["--analysis"] == "heuristic"):
            sys.exit(print("ERROR: WQFM works only in heuristic mode"))

    if arguments["convert"]:

        if not arguments["--filetype"] in ["trees","triplets"]:
            sys.exit(print("ERROR: --filetype flag must be one of: 'trees', " +
                           "'triplets'"))

    if arguments["convert"] or arguments["analysis"]:

        if not arguments["--analysis"] in ["auto","heuristic", "bandb"]:
            sys.exit(print("ERROR: --analysis flag must be one of: 'auto', " +
                           "'heuristic', 'bandb"))

        try:
            int(arguments["--replicates"])
        except ValueError:
            sys.exit(print("ERROR: --replicates flag must be an integer " +
                  "(number of replicates in heuristic search"))

    if arguments["tripdec"] or arguments["convert"] or arguments["analysis"]:

        if not arguments["--parallel"] in ["auto","no"]:
            try:
                int(arguments["--parallel"])
            except ValueError:
                sys.exit(print("ERROR: --parallel flag must be 'auto', 'no'" +
                      " or an integer specifying the number of cpu"))
                
        if arguments["--parallel"] != "no" and arguments["--detailed_tripdec"]:
            print("WARNING: detailed output for triplet decomposition " +
              "using the --detail flag is available only in " +
              "non-parallel mode.\nTo compute the detailed output file," +
              " please use --parallel=no.")

    if arguments["fp"] or arguments["analysis"]:
        if not arguments["--repetitions"] in ["TMS","FPS"]:
            sys.exit(print("ERROR: --repetitions flag must be one of: " +
                           "'TMS', 'FPS'"))

    if arguments["consensus"] or arguments["analysis"]:
        if arguments.get("--consensus", False):
            if not arguments["--consensus"] in ["strict","rcc"]:
                sys.exit(print("ERROR: --consensus flag must be one of: " +
                               "'strict', 'rcc'"))

    if arguments["support"]:
        if not arguments["--index"] in ["ri","tripdistance"]:
            sys.exit(print("ERROR: --index flag must be one of: 'ri', " +
                               "'tripdistance'"))


def helper(command):
    """
    Helper function that print the help block corresponding to a command.
    Parameters
    ----------
    option : str
        Any command that can be used with Agatta. Agatta commands comprises
        analysis, tripdec, ri, chartest, convert, fp, consensus, describetree,
        standardisation, hmatrix
    Returns
    -------
    None.
    """

    if command == "analysis":
        print("""
    analysis

    Main function of the Agatta python package. Allow to perform a three-item
    analysis, e.g., in the context of systematics phylogenetics, using
    hierarchical characters, or in cladistic biogeography. The cladogram(s)
    obtained by congruence is the tree that maximises the amount of hypotheses
    of cladistic relationships (i.e., the three-item statements) deduced from
    the input trees (the characters).
    The analysis can be performed using one or several text file containing a 
    hierarchical matrix (see section mandatory parameters below for 
    informations about the format), newick rooted trees encoded in a newick, 
    nexus, or nexml file, or a mix of input formats.
    There are no constraints on the input trees excepted that they must
    be rooted to perform the analysis. If they are repeated leaves
    (polymorphism), they are automatically removed (several methods
    are implemented, see below).
    Several options are available for analysing the results: the user can
    compute a consensus when several cladograms are optimal, a specific
    character-state testing procedure can be used to test whether each
    character state (i.e., each informative node of a tree) is a synapomorphy
    or an homoplasy, and finally a retention index can be computed to
    obtained the proportion of phylogenetic relationships of characters
    that have been retained in the optimal cladogram.

    Usage:

        agatta analysis <file>... [-s -v --analysis=<type> --chartest
                                      --consensus=<type> --parallel=<int>
                                      --pdf --prefix=<file>
                                      --repetitions=<type> --replicates=<int>
                                      --ri --rosetta=<file> --softpath=<path>
                                      --software=<type> --taxarep1=<path>
                                      --weighting=<type> --detailed_tripdec]

        Mandatory parameters:

            <file> Path(s) of the file(s) containing the character trees.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            Mixing input file format is allowed.
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --analysis=<type>  Type of tree search analysis between an exact
            branch and bound ('bandb') or an heuristic tree search
            ('heuristic'). The heuristic search is only available through
            PAUP*, TNT, WQFM, or wTREE-QMC thus it is mandatory to add the flag
            --software=tnt, --software=paup, --software=wqfm, or 
            software=wtree-qmc (and the flag --softpath accordingly).
            By default the analysis is in branch and bound below 15 terminals
            and heuristic otherwise.
            --chartest  Test and locates all character states on the cladogram.
            Each character state can be a synapomorphy if the hypothesis is
            accepted or an homoplasy if the hypothesis of sameness is rejected.
            Two output files are writen: prefix.chartest gives all locations
            of the states on the cladogram and if the test is passed or not,
            sorted by state, and prefix.chartest_node gives the same
            information sorted by node.
            --consensus=<type>  Compute a consensus which can be a strict
            consensus ('strict') or a reduced cladistic consensus ('rcc',
            Wilkinson 1994) which is able to detect more common information
            in subtrees. By default, the flag without argument produces a
            strict consensus. The output file is prefix.constrict or
            prefix.rcc.
            --parallel=<type>  Option for choosing if the analysis is made
            using multiprocessing or not. This argument can be:
              - 'not' if the user does not wan to use multiprocessing.
              - 'auto' for automatic detection of the number of cpu.
              - any integer corresponding to the number of cpu allowed.
            By default, the analysis is made in parallel using all available
            cpu. This option can be used in the case of very large character
            tree that can saturate the RAM if too many parallel processing are
            active.
            --pdf  Compute a pdf file to visualise character states on
            the cladogram if --chartest is used. One page for each state is
            writen. The file is named prefix.pdf.
            --prefix=<file>  Prefix of all saving files. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.
            --repetitions=<type>  The removal of repeated leaves in character
            trees is made using the method of free-paralogy subtree.
            Two algorithms are implemented, the original one from Nelson and
            Ladiges (1996) ('FPS') for dealing with paralogy in cladistic
            biogeography, and the algorithm of Rineau et al. (2021) ('TMS')
            designed for all cases of repetitions (not only paralogy).
            If the flag is not used and if repetitions are detected, they are
            automatically removed using Rineau et al. algorithm's.
            The repetition-free character trees are writen in prefix.poly and
            each new tree receives an id, e.g. 1.2 corresponds to the second
            repetition-free subtree computed from the 1st original character
            tree.
            --replicates=<int>  State the number of replicates in case
            of an heuristic tree search. By default the heuristic search is
            launched with 1000 replicates.
            --ri  Compute the retention index of the resulting cladogram and
            a retention index for each character which states the percentage
            of phylogenetic information retained in the optimal cladogram.
            The results are writen in prefix.ri.
            --rosetta=<file>  If the input tree leaves are parts, cladistics
            requires a standardisation step with replacement of parts to
            wholes. It is especially important in cladistic biogeography where
            terminal taxa are replaced by biogeographic areas. The rosetta flag
            replaces leaves of input trees according to a csv file with its
            path given as argument. The csv file is a table with two columns,
            one with the name of the tree leaves and the second with their
            corresponding names to be switched. The results of the
            standardisation are writen in the file prefix.stand.
            --softpath=<path>  Path of the software declared in --software.
            --software=<type>  Choose how to perform the three-item analysis.
            The analysis can be performed using the built-in branch and bound
            in Agatta ('agatta'). 'paup' and 'tnt' can be used for branch and
            bound or heuristic search. 'wqfm' and 'wtree-qmc' are conceived to 
            perform heuristic searches only.
            By default the analysis is made using built-in branch and bound.
            However it works only with very few terminals.
            User should consider to switch the software is the
            analysis time appears to be too long. A prefix.nex file is
            generated if 'paup', a prefix.tnt file for 'tnt', a
            prefix.wqfm for 'wqfm', and a prefix.wtqmc for 'wtree-qmc'.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.
            --weighting=<type>  Weighting scheme to use on triplets. The
            type of weighting will change the results of the analysis.
            The following schemes can be used:
                - FW: Fractional weighting from Rineau et al. (2021),
                - FWNL: Fractional weighting from Nelson and Ladiges (1992),
                - UW: Uniform weighting from Nelson and Ladiges (1992),
                - MW: Minimal weighting from Wilkinson et al. (2004),
                - AW: Additive weighting : the weight of a triplet in additive
                  weighting corresponds to the number of trees in which the
                  triplet is present,
                - NW: No weighting (all triplets have a weight of 1).
            By default 'FW' is used.
            --detailed_tripdec Compute a detailed csv table showing the
            link between triplet weights and character trees. Each column 
            corresponds to one character (same order as <file>). Each line 
            corresponds to a triplet. The last column and line give the sum
            of all weights of the column or line, respectively.
            --nsupport Compute node support for each resulting cladogram based 
            on triplets (amount weighted triplets compatible with each node). 
            When a consensus is computed, node support is also computed for
            them is nsupport option is activated.

    Output:

        Four ouput files are writen all the time when using the analysis
        command in addition to optionnal output files.
        The files are:
          - prefix.log is a log file with all the parameters of the analysis.
          - prefix.triplet is a file with all triplets deduced from the input
            character trees. Each row corresponds to one triplet with its
            weight as a fraction and as a float.
          - prefix.taxabloc is a table file with the correspondance between
            leaf identifiers and names given in the input.
          - prefix.tre is a newick file recording all the optimal cladograms
            found during the analysis.
              """)

    elif command == "tripdec":
        print("""
    tripdec

    Decomposes rooted tree(s) into minimal cladistic statements (triplets
    or three-item statements) stating that two leaves are closer between them
    than to a third. During decomposition the weight of each triplet is
    computed according to a specific weighting scheme. In three-item analysis,
    the weighted triplets are then analysed to compute the cladogram that is
    in agreement with the maximum amout of them.

    Usage:

        agatta tripdec <file>... [-s -v --parallel=<int> --prefix=<file> 
                                     --taxarep1=<path> --weighting=<type>
                                     --repetitions=<type> --detailed_tripdec]

        Mandatory parameters:

            <file> Path(s) of the file containing the character trees.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --parallel=<type>  Option for choosing if the analysis is made
            using multiprocessing or not. This argument can be:
              - 'not' if the user does not wan to use multiprocessing.
              - 'auto' for automatic detection of the number of cpu.
              - any integer corresponding to the number of cpu allowed.
            By default, the analysis is made in parallel using all available
            cpu. This option can be used in the case of very large character
            tree that can saturate the RAM if too many parallel processing are
            active.
            --prefix=<file>  Prefix of all saving files. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.
            --weighting=<type>  Weighting scheme to use on triplets. The
            type of weighting will change the results of the analysis.
            The following schemes can be used:
                - FW: Fractional weighting from Rineau et al. (2021),
                - FWNL: Fractional weighting from Nelson and Ladiges (1992),
                - UW: Uniform weighting from Nelson and Ladiges (1992),
                - MW: Minimal weighting from Wilkinson et al. (2004),
                - AW: Additive weighting : the weight of a triplet in additive
                  weighting corresponds to the number of trees in which the
                  triplet is present,
                - NW: No weighting (all triplets have a weight of 1).
            By default 'FW' is used.
            --detailed_tripdec Compute a detailed csv table showing the
            link between triplet weights and character trees. Each column 
            corresponds to one character (same order as <file>). Each line 
            corresponds to a triplet. The last column and line give the sum
            of all weights of the column or line, respectively.
            --repetitions=<type>  The removal of repeated leaves in character
            trees is made using the method of free-paralogy subtree.
            Two algorithms are implemented, the original one from Nelson and
            Ladiges (1996) ('FPS') for dealing with paralogy in cladistic
            biogeography, and the algorithm of Rineau et al. (2021) ('TMS')
            designed for all cases of repetitions (not only paralogy).
            If the flag is not used and if repetitions are detected, they are
            automatically removed using Rineau et al. algorithm's.
            The repetition-free character trees are writen in prefix.poly and
            each new tree receives an id, e.g. 1.2 corresponds to the second
            repetition-free subtree computed from the 1st original character
            tree.

    Output:

        Two files are writen after decomposition. The main one is
        prefix.triplet, a file containing all triplets deduced from the input
        character trees. Each row corresponds to one triplet with its
        weight as an integer (if UW, AW, or NW), or as a fraction and as a
        float otherwise. The second file prefix.taxabloc is a table file with
        the correspondance between leaf identifiers and names given in the
        input trees.
              """)

    elif command == "support":
        print("""
    support

    The support command gathers several triplet metrics and indices that are
    intended to compare trees between them. The retention index measures the
    amount of cladistic relationships from the input characters retained in
    the optimal cladogram (or a consensus). The triplet distance measures
    the distance in terms of triplets between two trees. The ITRI compares
    in terms of triplets a tree relatively to a reference tree (used to
    measure efficiency of methods using simulations).
    All these metrics handle weighting schemes.

    Usage:

        agatta support <file> <file>... [-s -v --index=<type> --prefix=<file>
                                       --taxarep1=<path> --taxarep2=<path>
                                       --weighting=<type> --repetitions=<type>]

        Mandatory parameters:

            At least two <file> arguments are requested which represents path 
            of tree files. The first is the cladogram, the others are 
            character files.
            The requested files depend of the --index flag:
                --index=ri: the retention index compares the cladogram to its
                characters. The first <file> contains one tree considered as
                the optimal cladogram (or a consensus); the second <file>
                contains the input character trees used to construct the
                cladogram and can be newick files or a hierarchical matrix.
                --index=tripdistance: compute various measures for comparison 
                between only two trees t1 and t2 (one in each file). 

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --index  The index flag specifies which measure to use to compare
            rooted trees between retention index ('ri'), or inter-tree 
            retention index/triplet distance ('tripdistance'). More 
            informations on the input file in the mandatory parameters section.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.
            --taxarep2=<path>  Idem as --taxarep1 for the second <file>.
            --weighting=<type>  Weighting scheme to use on triplets. The
            type of weighting will change the results of the analysis.
            The following schemes can be used:
                - FW: Fractional weighting from Rineau et al. (2021),
                - FWNL: Fractional weighting from Nelson and Ladiges (1992),
                - UW: Uniform weighting from Nelson and Ladiges (1992),
                - MW: Minimal weighting from Wilkinson et al. (2004),
                - AW: Additive weighting : the weight of a triplet in additive
                  weighting corresponds to the number of trees in which the
                  triplet is present,
                - NW: No weighting (all triplets have a weight of 1).
            By default 'FW' is used.
            --repetitions=<type> The removal of repeated leaves in character
            trees is made using the method of free-paralogy subtree.
            Two algorithms are implemented, the original one from Nelson and
            Ladiges (1996) ('FPS') for dealing with paralogy in cladistic
            biogeography, and the algorithm of Rineau et al. (2021) ('TMS')
            designed for all cases of repetitions (not only paralogy).
            If the flag is not used and if repetitions are detected, they are
            automatically removed using Rineau et al. algorithm's.
            The repetition-free character trees are writen in prefix.poly and
            each new tree receives an id, e.g. 1.2 corresponds to the second
            repetition-free subtree computed from the 1st original character
            tree. 

    Output:

        One ouput file is writen the results of the tree comparisons:
            - For ri, a global retention index stating the overall information
            content of all character trees retained in the cladogram is writen,
            plus a ri for each character, a ri for each character state, and a 
            ri for each subtree if polymorphism allowing to cut tcharacter 
            trees in subtrees present.
            - For tripdistance, several values are computed given two trees t1 
            (or reference/true tree) and t2 (or reconstructed tree) ('number of 
            triplets' can refer to a sum of triplet weights depending of the
            weighting scheme chosen). The interpetation differs if one wants to 
            compare two equal topologies or if the comparison involves a 
            reference tree and a tree to be compared with (in this case, the 
            significance is added in parentheses):
                
                * Number of triplets in t1 (relevant elements)
                
                * Number of triplets in t2 (retreived elements)
                
                * Number of triplets both in t1 and t2 (true positives)
                  Note that t1 triplets present in t2 and t2 triplets also in 
                  t1 may differ because of the weighting)
            
                * Number of triplets in t2 but not in t1 (false positives) 
                
                * Number of triplets in t1 but not in t2 (false negatives) 
                
                * ITRI(t1,t2) (Precision: (t2 triplets present in t1)/t2) 
                  amount of triplets from the reconstructed tree that are true.) 
            
                * ITRI(t2,t1) (Recall: (t1 triplets present in t2)/t1) amount 
                  of true triplets that are present in the reconstructed tree) 
            
                * Triplet distance (F1-score: harmonic mean of precision and 
                  recall (2 * Precision * Recall) / (Precision + Recall))

        All calculations are made using a specific weighting scheme
        (option --weighting).
              """)

    elif command == "chartest":
        print("""
    chartest

    The chartest command is intended to use the character state testing
    procedure for hierarchical characters. Each character state
    (an informative node in a rooted tree) is tested against a cladogram:
    if the test fails, the character state hypothesis (an hypothesis of clade)
    is rejected, and the state becomes homoplasic; otherwise, if the state pass
    the test, it becomes a synapomorphy that supports a specific node of the
    cladogram.

    Usage:

        agatta chartest <file> <file>... [-s -v --pdf --prefix=<file> 
                                       --taxarep1=<path> --taxarep2=<path> 
                                       --repetitions=<type>]

        Mandatory parameters:

            <file> the first <file> contains one tree considered as
            the optimal cladogram (or a consensus);  the others are the path(s) 
            of file(s) containing the character trees.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.
            The first <file> contains only one tree: the optimal cladogram or
            the consensus tree resulting from the analysis of a set of
            character trees. The second <file> contains the set of character
            trees.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --pdf  Compute a pdf file to visualise character states on
            the cladogram if --chartest is used. One page for each state is
            writen. The file is named prefix.pdf.
            --prefix=<file>  Prefix of all saving files. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.
            --taxarep2=<path>  Idem as --taxarep1 for the second <file>.
            --repetitions=<type>  The removal of repeated leaves in character
            trees is made using the method of free-paralogy subtree.
            Two algorithms are implemented, the original one from Nelson and
            Ladiges (1996) ('FPS') for dealing with paralogy in cladistic
            biogeography, and the algorithm of Rineau et al. (2021) ('TMS')
            designed for all cases of repetitions (not only paralogy).
            If the flag is not used and if repetitions are detected, they are
            automatically removed using Rineau et al. algorithm's.
            The repetition-free character trees are writen in prefix.poly and
            each new tree receives an id, e.g. 1.2 corresponds to the second
            repetition-free subtree computed from the 1st original character
            tree.
    Output:

        Two output files are writen: prefix.chartest gives all locations
        of the states on the cladogram and if the test is passed or not,
        sorted by state, and prefix.chartest_node gives the same
        information sorted by node. For better visualisation, the flag --pdf
        allow to generate a pdf file with a page for each character state with
        the resulting location and results.
              """)

    elif command == "convert":
        print("""
    convert

    The convert command is intended to compute a triplet matrix readable by an
    external software (currently PAUP*, TNT, WQFM, wTREE-QMC are implemented) 
    from a file containing a hierarchical matrix, a list of trees, or a list of 
    triplets.

    Usage:

        agatta convert <file>... [-s -v --analysis=<type> --filetype=<type> 
                                      --log --parallel=<int> --prefix=<file>
                                      --replicates=<int> --software=<type>
                                      --taxarep1=<path> --weighting=<type>
                                      --repetitions=<type>]

        Mandatory parameters:

            <file> Path(s) of the file(s) containing character trees or 
            triplets.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.
            The user can also use as input file a triplet file generated from
            the Agatta tripdec command.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --analysis=<type>  Type of tree search analysis between an exact
            branch and bound ('bandb') or an heuristic tree search
            ('heuristic'). A line is writen accordingly in the output file.
            By default the analysis is in branch and bound below 15 terminals
            and heuristic otherwise.
            --filetype=<type>  The input file can be a classic file containing
            newick trees (newick file, nexus file, or nexml file) ('trees'), or
            a list of triplets ('triplets') generated using the Agatta tripdec
            command.
            --log  Add a line in the output file to generate a log file during
            the PAUP* or TNT analysis.
            --parallel=<type>  Option for choosing if the triplet decomposition
            is made using multiprocessing or not. This argument can be:
              - 'no' if the user does not want to use multiprocessing.
              - 'auto' for automatic detection of the number of cpu.
              - any integer corresponding to the number of cpu allowed.
            By default, the analysis is made in parallel using all available
            cpu. This option can be used in the case of very large character
            tree that can saturate the RAM if too many parallel processing are
            active.
            --prefix=<file>  Prefix of saving .nex or .tnt file. The complete
            path can be used. By default, the prefix is 'agatta_out' and output
            file is saved in the directory of <file>.
            --replicates=<int>  State the number of replicates in case
            of an heuristic tree search. By default the heuristic search is
            launched with 1000 replicates.
            --software=<type>  Choose how to perform the three-item analysis.
            The analysis can be performed using the built-in branch and bound
            in Agatta ('agatta'). 'paup' and 'tnt' can be used for branch and
            bound or heuristic search. 'wqfm' and 'wtree-qmc' can be used for 
            heuristic search only. By default the analysis is made using
            built-in branch and bound. However it works only with very few
            terminals. User should consider to switch the software is the
            analysis time appears to be too long. A prefix.nex file is
            generated if 'paup', a prefix.tnt file for 'tnt', a prefix.wqfm
            for 'wqfm', and a prefix.wtqmc for 'wtree-qmc'.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.
            --weighting=<type>  Weighting scheme to use on triplets. The
            type of weighting will change the results of the analysis.
            The following schemes can be used:
                - FW: Fractional weighting from Rineau et al. (2021),
                - FWNL: Fractional weighting from Nelson and Ladiges (1992),
                - UW: Uniform weighting from Nelson and Ladiges (1992),
                - MW: Minimal weighting from Wilkinson et al. (2004),
                - AW: Additive weighting : the weight of a triplet in additive
                  weighting corresponds to the number of trees in which the
                  triplet is present,
                - NW: No weighting (all triplets have a weight of 1).
            By default 'FW' is used.
            --repetitions=<type>  The removal of repeated leaves in character
            trees is made using the method of free-paralogy subtree.
            Two algorithms are implemented, the original one from Nelson and
            Ladiges (1996) ('FPS') for dealing with paralogy in cladistic
            biogeography, and the algorithm of Rineau et al. (2021) ('TMS')
            designed for all cases of repetitions (not only paralogy).
            If the flag is not used and if repetitions are detected, they are
            automatically removed using Rineau et al. algorithm's.
            The repetition-free character trees are writen in prefix.poly and
            each new tree receives an id, e.g. 1.2 corresponds to the second
            repetition-free subtree computed from the 1st original character
            tree.

    Output:

        A triplet matrix file with weights that can be analysed using
        PAUP* or TNT or a triplet file with weights for WQFM or wTREE-QMC.
              """)

    elif command == "fp":
        print("""
    fp

    The free-paralogy analysis is a method used to manage repetitions
    (polymorphism in phylogenetics) in the context of cladistic analysis
    using hierarchical characters. Free-paralogy subtree analysis build
    subtrees to avoid repetition of leaves. Two distinct algorithm for building
    subtrees are currently implemented in Agatta (see --repetitions).

    Usage:

        agatta fp <file>... [-s -v --prefix=<file> --repetitions=<type>
                                 --taxarep1=<path>]

        Mandatory parameters:

            <file> Path(s) of the file containing the character trees.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --prefix=<file>  Prefix of all saving files. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.
            --repetitions=<type>  Two algorithms are implemented, the original
            one from Nelson and Ladiges (1996) ('FPS') for dealing with
            paralogy in cladistic biogeography, and the algorithm of
            Rineau et al. (2021) ('TMS') designed for all cases of
            repetitions (not only paralogy).
            If the flag is not used and if repetitions are detected, they are
            automatically removed using Rineau et al. algorithm's.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.

    Output:

        One newick file prefix.poly with a list of subtrees free of repeated
        leaves. Each subtree is labelled with the number of the original tree
        and the number of the subtree generated from the original tree,
        e.g. 1.2 corresponds to the second repetition-free subtree computed
        from the 1st original character tree.
              """)

    elif command == "consensus":
        print("""
    consensus

    Compute a consensus of several equally optimal trees. Two types of
    consensus are available: a strict consensus, that displays only the clades
    common to all trees, and a reduced cladistic consensus (rcc), that displays
    subtrees that are common to all trees (or in other words, the rcc
    generates all triplets common to all optimal trees and combines them
    into the bigest trees).

    Usage:

        agatta consensus <file> [-s -v --consensus=<type> --prefix=<file>
                                       --taxarep1=<path>]
        Mandatory parameters:

            <file> Path of the file containing the equally optimal trees.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --consensus=<type>  Compute a consensus which can be a strict
            consensus ('strict') or a reduced cladistic consensus ('rcc',
            Wilkinson 1994) which is able to detect more common information
            in subtrees. By default, the flag without argument produces a
            strict consensus.
            --prefix=<file>  Prefix of all saving files. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.
            --taxarep1=<path>  If the user wants to replace identifiers by real
            leaf names in the result files, this flag can be used with a path
            to a csv file with two columns, the first with the identifiers in
            the actual newick strings and the other with the names the user
            wants.

    Output:

        One newick file containing the strict consensus prefix.constrict or
        a list of common subtrees (the rcc profile) prefix.rcc.
              """)

    elif command == "describetree":
        print("""
    describetree

    The describetree command output basic informations on a list of
    rooted trees. The informations writen are related to the number of
    nodes, terminals, internal nodes, symmetric nodes, apical nodes, to the
    sesolution of the tree, number of dichotomies, polytomies.

    Usage:

        agatta describetree <file> [-s -v --prefix=<file> --showtaxanames]

        Mandatory parameters:

            <file> Path of the file containing rooted trees.
            The trees can be encoded for Agatta in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            THe following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --prefix=<file>  Prefix of the saving file. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.
            --showtaxanames  The list of all tree leaves is writen in the
            output file.

    Output:

        One output file prefix.dt with descriptors for each tree.
              """)

    elif command == "standardisation":
        print("""
    standardisation

    This command is used for standardisation of characters in the context of
    cladistic biogeography, i.e., replacement of leaves using a correspondence
    table. In cladistic theory, the standardisation is the construction of
    character trees (hypotheses of kinship relationship between bearers) based
    on homologies (hypotheses of kinship relationships between parts). The
    standardisation is used to convert phylogenies in areagrams in cladistic
    biogeography. The only currently implemented option for managing MAST is
    their automatic deletion (i.e. MAST do not bear any unambiguous
    biogeographic information).

    Usage:

        agatta standardisation <file> <file>... [-s -v --prefix=<file>]

        Mandatory parameters:

            Two <file> arguments at least are requested with the 
            standardisation command. 
            The first <file> argument is the path of a csv table with two
            columns, the first column corresponds to the leaf names of input
            trees, and the second column to corresponding names for
            replacement. For example, in cladistic biogeography, taxa are in
            the left column and corresponding areas in the right column.
            
            The other(s) <file> correspond to the path(s) of the file(s) 
            containing the character trees. The trees can be encoded for Agatta 
            in a file in several ways:
              - a hierarchical matrix,
              - A newick file with a single newick tree on each line,
              - A nexus file (extension in .nex),
              - A nexml file (extension in .nexml).
            The following url give more information on how to build input
            files: https://vrineau.github.io/AgattaDocs/Input%20files.html.


        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --prefix=<file>  Prefix of the saving files. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.

    Output:

        One output newick file prefix.stand containing the trees after
        leaf replacement.
              """)

    elif command == "hmatrix":
        print("""
    hmatrix

    hmatrix converts one or several hierarchical matrices into rooted trees. 
    Each column of the matrix corresponds to one tree.

    Usage:

        agatta analysis <file>... [-s -v --chardec --prefix=<file>]

        Mandatory parameters:

            <file>  Hierarchical matrix (one or several). 
            A complete guide on the hierarchical matrix format is available 
            here: 
            https://vrineau.github.io/AgattaDocs/Input%20files.html.

        Optionnal parameters:

            -s  Silent mode.
            -v  Verbose mode.
            --chardec  Decompose each tree into components (one subtree for
            each informative node).
            --prefix=<file>  Prefix of the saving file. The complete path can
            be used. By default, the prefix is 'agatta_out' and all files are
            saved in the directory of the first <file>.

    Output:

        One output file prefix.tre with the trees deduced from the hierarchical
        matrix.
              """)

    else:
        print("""This command does not exist. Available commands:
              analysis, tripdec, support, chartest, convert, fp, consensus,
              describetree, standardisation, hmatrix.""")

    sys.exit(1)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path('agatta/__version__.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setup(

    name='Agatta',

    version=main_ns['__version__'],

    python_requires='>=3.8.0',

    packages=find_packages(),

    author="Valentin Rineau",

    author_email="valentin.rineau@sorbonne-universite.fr",

    description="Three-item analysis python package",

    long_description_content_type="text/markdown",

    long_description=open('README.md').read(),

    install_requires= ["docopt==0.6.2",
                       "ete3==3.1.3",
                       "tqdm==4.66.3",
                       "treeswift==1.1.37",
                       "six==1.16.0",
                       "numpy==1.24.4",
                       "pandas==2.2.3",
                       "pypdf==3.17.0"],

    include_package_data=True,

    url='https://github.com/vrineau/Agatta',

    entry_points = {
        'console_scripts': [
            'agatta = agatta:main',
        ],
    },

    license="GNU General Public License v3.0+",

    classifiers = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: MacOS",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Education",
    "Intended Audience :: Other Audience",
    "Natural Language :: English",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Visualization",
]

)

